class InputExample:
    """
    A single training/test example for simple sequence classification
    """

    def __init__(self, guid: str, text: str, tags: str = ""):
        """
        :param guid: [str] unique id for input example
        :param text: [str] raw (untokenized) text of first sequence.
        :param tags: [str] labels of first sequence, separated by whitespace.
        """
        self.guid = guid
        self.text = text
        self.tags = tags



import argparse
from nerblackbox.modules.datasets.formatter.auto_formatter import AutoFormatter
import mlflow


def main(args):
    """
    - I: get data for ner_dataset
    - II: write ner_tag_mapping.json file
    - III: format data
    - IV: resplit data
    - V: analyze and plot data
    --------------------------------------------------------------------------------
    :return: -
    """
    with mlflow.start_run(run_name="Default"):
        # formatter
        formatter = AutoFormatter.for_dataset(args.ner_dataset)

        formatter.create_directory()
        formatter.get_data(verbose=args.verbose)  # I: get_data
        formatter.create_ner_tag_mapping_json(
            modify=args.modify
        )  # II: create ner tag mapping
        formatter.format_data()  # III: format data
        formatter.resplit_data(val_fraction=args.val_fraction)  # IV: resplit data
        formatter.analyze_data()  # V: analyze data
        formatter.plot_data()  # V: analyze data


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ner_dataset", required=True, type=str, help="e.g. swedish_ner_corpus"
    )
    parser.add_argument("--modify", type=bool, default=True)
    parser.add_argument("--val_fraction", type=float, default=0.3)
    parser.add_argument("--verbose", type=bool, default=False)
    _args = parser.parse_args()

    main(_args)



import matplotlib.pylab as plt


class Plots:

    clr = ["k", "g", "orange", "r"]
    shift = [-0.3, -0.1, 0.1, 0.3]
    width = 0.1
    column_names = ["tags/sentence", "tags relative w/ 0", "tags relative w/o 0"]

    def __init__(self, _stats_aggregated, _num_sentences):
        self.stats_aggregated = _stats_aggregated
        self.num_sentences = _num_sentences
        self.num_tokens = None
        self.tags = list(self.stats_aggregated["total"].index)

    def plot(self, fig_path=None):
        ################################################################################################################
        # 3. START
        ################################################################################################################
        phases = ["train", "val", "test"]
        phases_all = ["total"] + phases

        self.num_tokens = {
            phase: self.get_tokens(self.stats_aggregated[phase]) for phase in phases_all
        }
        columns = {
            column_name: {
                phase: self.get_columns(self.stats_aggregated[phase], column_name)
                for phase in phases_all
            }
            for column_name in self.column_names
        }

        fig, ax = plt.subplots(2, 2, figsize=(12, 8))
        ax0, ax1, ax2, ax3 = ax.flatten()
        self.plot_sentences_and_tokens(ax0, self.num_sentences, self.num_tokens)
        self.plot_column(
            ax1,
            columns,
            "tags/sentence",
            legend={"loc": "upper right", "bbox_to_anchor": (1.24, 1.0)},
        )
        self.plot_column(ax2, columns, "tags relative w/ 0", y_normalize=True)
        self.plot_column(ax3, columns, "tags relative w/o 0", y_normalize=True)

        if fig_path is not None:
            print("\nsave figure at", fig_path)
            plt.savefig(fig_path)

    ####################################################################################################################
    # GET DATA FROM DF
    ####################################################################################################################
    @staticmethod
    def get_tokens(df):
        return df.loc[:, "tags"].sum()

    @staticmethod
    def get_columns(df, column_name: str):
        return df.loc[:, column_name].to_dict()

    ####################################################################################################################
    # PLOT SENTENCES AND TOKENS
    ####################################################################################################################
    def plot_sentences_and_tokens(self, _ax, _sentences, _tokens, legend=False):
        tags = ["sentences", "tokens"]
        xs = list(range(len(tags)))
        labels = list(_sentences.keys())

        def normalize(_list):
            return [elem / _list[0] for elem in _list], int(_list[0])

        values_sentences, total_sentences = normalize(list(_sentences.values()))
        values_tokens, total_tokens = normalize(list(_tokens.values()))

        for n, tag in enumerate(tags):
            x = xs[n]
            for i in range(len(self.shift)):
                if n == 0:
                    _ax.bar(
                        x + self.shift[i],
                        values_sentences[i],
                        label=labels[i] if n == 0 else None,
                        width=self.width,
                        color=self.clr[i],
                    )
                else:
                    _ax.bar(
                        x + self.shift[i],
                        values_tokens[i],
                        label=labels[i] if n == 0 else None,
                        width=self.width,
                        color=self.clr[i],
                    )
            if legend:
                _ax.legend()
            _ax.set_xticks(xs)
            _ax.set_xticklabels(tags)
            _ax.set_ylim([0, 1])
            _ax.set_yticks([1])
            _ax.set_yticklabels([total_sentences])

            _ax2 = _ax.twinx()
            _ax2.set_ylim([0, 1])
            _ax2.set_yticks([1])
            _ax2.set_yticklabels([total_tokens])

    def plot_column(self, _ax, _columns, column_name, y_normalize=False, legend=None):
        column = _columns[column_name]

        xs = list(range(len(self.tags)))
        labels = list(column.keys())

        for n, tag in enumerate(self.tags):
            x = xs[n]
            for i in range(len(self.shift)):
                _ax.bar(
                    x + self.shift[i],
                    float(column[labels[i]][tag]),
                    label=labels[i] if n == 0 else None,
                    width=self.width,
                    color=self.clr[i],
                )
            if n == len(self.tags) - 1:
                if legend is not None:
                    _ax.legend(**legend)
                _ax.set_xticks(xs)
                _ax.set_xticklabels(self.tags)
        _ax.set_ylabel(column_name)
        if y_normalize:
            _ax.set_ylim([0, 1])



from dataclasses import dataclass
from dataclasses import asdict
from typing import List, Optional

from sklearn.metrics import accuracy_score as accuracy_sklearn
from sklearn.metrics import precision_score as precision_sklearn
from sklearn.metrics import recall_score as recall_sklearn
from sklearn.metrics import precision_recall_fscore_support as prf_sklearn
from sklearn.exceptions import UndefinedMetricWarning

import warnings

from seqeval.metrics import precision_score as precision_seqeval
from seqeval.metrics import recall_score as recall_seqeval
from seqeval.metrics import f1_score as f1_seqeval


class NerMetrics:
    def __init__(
        self,
        true_flat,
        pred_flat,
        tag_list=None,
        level="token",
        plain_tags=False,
        verbose=False,
    ):
        """
        :param true_flat: [np array] of shape [batch_size * seq_length]
        :param pred_flat: [np array] of shape [batch_size * seq_length]
        :param tag_list:  [optional, list] of [str] labels to take into account for metrics
        :param level:     [optional, str] 'token' or 'chunk'
        :param verbose:   [optional, bool] if True, show verbose output
        """
        self.true_flat = true_flat
        self.pred_flat = pred_flat
        self.tag_list = tag_list
        self.level = level
        self.verbose = verbose

        self.results = Results()
        self.failure_value = -1

        if self.level == "chunk":
            self.true_flat_bio = convert_to_chunk(self.true_flat, to_bio=plain_tags)
            self.pred_flat_bio = convert_to_chunk(self.pred_flat, to_bio=plain_tags)

    def results_as_dict(self):
        return asdict(self.results)

    def compute(self, _metrics):
        """
        computes selected metrics
        ----------------------------------------------------------
        :param _metrics: [list] of [str], e.g. ['acc, 'precision']
        :return: -
        """
        warnings.filterwarnings("error")

        if "acc" in _metrics:
            self.accuracy()
        if "precision" in _metrics:
            self.precision()
        if "recall" in _metrics:
            self.recall()
        if "f1" in _metrics:
            self.f1_score()

        warnings.resetwarnings()

    def accuracy(self):
        """
        computes accuracy of predictions (_np_logits) w.r.t. ground truth (_np_label_ids)
        ---------------------------------------------------------------------------------
        :return: acc [np float]
        """
        self.results.acc = accuracy_sklearn(
            self.true_flat, self.pred_flat, normalize=True
        )

    def precision(self):
        """
        computes precision (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)
        -----------------------------------------------------------------------------------------------
        :return: precision_macro [np array] for each class, then averaged
                 precision_micro [np array] for all examples
        """
        if self.level == "token":
            # precision_macro
            try:
                self.results.precision_macro = precision_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="macro",
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.precision_macro = self.failure_value

            # precision_micro
            try:
                self.results.precision_micro = precision_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="micro",
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.precision_micro = self.failure_value
        else:
            self.results.precision_micro = precision_seqeval(
                self.true_flat_bio, self.pred_flat_bio, average="micro"
            )

    def recall(self):
        """
        computes recall (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)
        -----------------------------------------------------------------------------------------------
        :return: recall_macro [np array] for each class, then averaged
                 recall_micro [np array] for all examples
        """
        if self.level == "token":
            try:
                # recall_macro
                self.results.recall_macro = recall_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="macro",
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.recall_macro = self.failure_value

            # recall_micro
            try:
                self.results.recall_micro = recall_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="micro",
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.recall_micro = self.failure_value
        else:
            self.results.recall_micro = recall_seqeval(
                self.true_flat_bio, self.pred_flat_bio, average="micro"
            )

    def f1_score(self):
        """
        computes f1 score (macro/micro) of predictions (_pred_flat) w.r.t. ground truth (_true_flat)
        -----------------------------------------------------------------------------------------------
        :return: f1_score_macro [np array] for each class, then averaged
                 f1_score_micro [np array] for all examples
        """
        if self.level == "token":
            try:
                # f1_macro
                _, _, self.results.f1_macro, _ = prf_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="macro",
                    warn_for=("precision", "recall", "f-score"),
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.f1_macro = self.failure_value

            # f1_micro
            try:
                _, _, self.results.f1_micro, _ = prf_sklearn(
                    self.true_flat,
                    self.pred_flat,
                    labels=self.tag_list,
                    average="micro",
                    warn_for=("precision", "recall", "f-score"),
                )
            except UndefinedMetricWarning as e:
                if self.verbose:
                    print(e)
                self.results.f1_micro = self.failure_value
        else:
            self.results.f1_micro = f1_seqeval(
                self.true_flat_bio, self.pred_flat_bio, average="micro"
            )


@dataclass
class Results:
    acc: float = -1
    precision_micro: float = -1
    precision_macro: float = -1
    recall_micro: float = -1
    recall_macro: float = -1
    f1_micro: float = -1
    f1_macro: float = -1


def convert_to_chunk(tag_list: List[str], to_bio=True) -> List[str]:
    """
    - get rid of special tokens
    - add bio prefixed to tags

    Args:
        tag_list:      e.g. ['O',   'ORG',   'ORG']
        to_bio:        whether to cast to bio labels

    Returns:
        bio_tag_list:  e.g. ['O', 'B-ORG', 'I-ORG']
    """
    clean_tag_list = get_rid_of_special_tokens(tag_list)
    if to_bio:
        assert_plain_tags(clean_tag_list)
        return add_bio_to_tag_list(clean_tag_list)
    else:
        assert_bio_tags(clean_tag_list)
        return clean_tag_list


def assert_plain_tags(tag_list: List[str]) -> None:
    for tag in tag_list:
        if tag != "O" and (len(tag) > 2 and tag[1] == "-"):
            raise Exception(
                "ERROR! attempt to convert tags to bio format that already seem to have bio format."
            )


def assert_bio_tags(tag_list: List[str]) -> None:
    for tag in tag_list:
        if tag != "O" and (len(tag) <= 2 or tag[1] != "-"):
            raise Exception(
                "ERROR! assuming tags to have bio format that seem to have plain format instead."
            )


def add_bio_to_tag_list(tag_list: List[str]) -> List[str]:
    """
    adds bio prefixes to tags

    Args:
        tag_list:     e.g. ['O',   'ORG',   'ORG']

    Returns:
        bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']
    """
    return [
        _add_bio_to_tag(tag_list[i], previous=tag_list[i - 1] if i > 0 else None)
        for i in range(len(tag_list))
    ]


def _add_bio_to_tag(tag: str, previous: Optional[str] = None) -> str:
    """
    add bio prefix to tag, depending on previous tag

    Args:
        tag:      e.g. 'ORG'
        previous: e.g. 'ORG'

    Returns:
        bio_tag:  e.g. 'I-ORG'
    """
    if tag == "O" or tag.startswith("["):
        return tag
    elif previous is None:
        return f"B-{tag}"
    elif tag != previous:
        return f"B-{tag}"
    else:
        return f"I-{tag}"


def get_rid_of_special_tokens(tag_list: List[str]) -> List[str]:
    """
    replace special tokens ('[CLS]', '[SEP]', '[PAD]') by 'O'

    Args:
        tag_list:         e.g. ['[CLS]', 'O', 'ORG', 'ORG', '[SEP]']

    Returns:
        cleaned_tag_list: e.g. [    'O', 'O', 'ORG', 'ORG',     'O']
    """

    return [tag if not tag.startswith("[") else "O" for tag in tag_list]



import os
import subprocess
from os.path import join

from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter
from nerblackbox.modules.utils.env_variable import env_variable


class SwedishNerCorpusFormatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "swedish_ner_corpus"
        ner_tag_list = ["PER", "ORG", "LOC", "MISC", "PRG", "ORG*"]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool):
        """
        I: get data
        -----------
        :param verbose: [bool]
        :return: -
        """
        bash_cmds = [
            f'git clone https://github.com/klintan/swedish-ner-corpus.git {env_variable("DIR_DATASETS")}/_swedish_ner_corpus',
            f'mv {env_variable("DIR_DATASETS")}/_swedish_ner_corpus/* {env_variable("DIR_DATASETS")}/swedish_ner_corpus',
            f'rm -rf {env_variable("DIR_DATASETS")}/_swedish_ner_corpus',
        ]

        for bash_cmd in bash_cmds:
            if verbose:
                print(bash_cmd)

            try:
                subprocess.run(bash_cmd, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                print(e)

    def create_ner_tag_mapping(self):
        """
        II: customize ner_training tag mapping if wanted
        -------------------------------------
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return {
            "ORG*": "ORG",
            "PRG": "O",
        }

    def format_data(self):
        """
        III: format data
        ----------------
        :return: -
        """
        for phase in ["train", "test"]:
            rows = self._read_original_file(phase)
            self._write_formatted_csv(phase, rows)

    def resplit_data(self, val_fraction: float):
        """
        IV: resplit data
        ----------------
        :param val_fraction: [float], e.g. 0.3
        :return: -
        """
        # train -> train, val
        df_train_val = self._read_formatted_csvs(["train"])
        df_train, df_val = self._split_off_validation_set(df_train_val, val_fraction)
        self._write_final_csv("train", df_train)
        self._write_final_csv("val", df_val)

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])
        self._write_final_csv("test", df_test)

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    def _read_original_file(self, phase):
        """
        - read original text files
        ---------------------------------------------
        :param phase:   [str] 'train' or 'test'
        :return: _rows: [list] of [list] of [str], e.g. [['Inger', 'PER'], ['säger', '0'], ..]
        """
        file_path_original = join(self.dataset_path, f"{phase}_corpus.txt")

        _rows = list()
        if os.path.isfile(file_path_original):
            with open(file_path_original) as f:
                for row in f.readlines():
                    _rows.append(row.strip().split())
            print(f"\n> read {file_path_original}")

        _rows = [row if len(row) == 2 else list() for row in _rows]

        return _rows



import pytest
import torch
from transformers import AutoTokenizer
from typing import List, Dict
from pkg_resources import resource_filename
from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import (
    CsvReader,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_examples_to_tensors import (
    InputExamplesToTensors,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.bert_dataset import (
    BertDataset,
)
from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
    DataPreprocessor,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
    EncodingsKeys,
)
from nerblackbox.tests.utils import PseudoDefaultLogger
from nerblackbox.modules.ner_training.ner_model import NEWLINE_TOKENS

tokenizer = AutoTokenizer.from_pretrained(
    "af-ai-center/bert-base-swedish-uncased",
    do_lower_case=False,
    additional_special_tokens=NEWLINE_TOKENS,
    use_fast=True,
)

csv_reader = CsvReader(
    path=resource_filename("nerblackbox", "tests/test_data"),
    tokenizer=tokenizer,
    do_lower_case=False,
    csv_file_separator="\t",
    default_logger=None,
)

data_preprocessor = DataPreprocessor(
    tokenizer=tokenizer,
    do_lower_case=False,
    default_logger=PseudoDefaultLogger(),
    max_seq_length=4,
)


class TestCsvReaderAndDataProcessor:

    ####################################################################################################################
    @pytest.mark.parametrize(
        "tag_list, " "input_examples",
        [
            (
                ["O", "PER", "ORG", "MISC"],
                {
                    "train": [
                        InputExample(
                            guid="",
                            text="På skidspår.se kan längdskidåkare själva betygsätta förhållandena i spåren .",
                            tags="O MISC O O O O O O O O",
                        ),
                    ],
                    "val": [
                        InputExample(
                            guid="",
                            text="Fastigheten är ett landmärke designad av arkitekten Robert Stern .",
                            tags="O O O O O O O PER PER O",
                        ),
                    ],
                    "test": [
                        InputExample(
                            guid="",
                            text="Apple noteras för 85,62 poäng , vilket är den högsta siffran någonsin i undersökningen .",
                            tags="ORG O O O O O O O O O O O O O O",
                        ),
                    ],
                    "predict": [
                        InputExample(
                            guid="",
                            text="På skidspår.se kan längdskidåkare själva betygsätta förhållandena i spåren .",
                            tags="O MISC O O O O O O O O",
                        ),
                        InputExample(
                            guid="",
                            text="Fastigheten är ett landmärke designad av arkitekten Robert Stern .",
                            tags="O O O O O O O PER PER O",
                        ),
                        InputExample(
                            guid="",
                            text="Apple noteras för 85,62 poäng , vilket är den högsta siffran någonsin i undersökningen .",
                            tags="ORG O O O O O O O O O O O O O O",
                        ),
                    ],
                },
            ),
        ],
    )
    def tests(
        self,
        tag_list: List[str],
        input_examples: Dict[str, List[InputExample]],
    ) -> None:

        ##################################
        # 1. CsvReader
        ##################################
        for phase in ["train", "val", "test"]:
            test_input_examples = csv_reader.get_input_examples(phase)
            assert (
                len(test_input_examples) == 1 or 2
            ), f"ERROR! len(test_input_examples) = {len(test_input_examples)} should be 1 or 2."
            assert (
                test_input_examples[0].text == input_examples[phase][0].text
            ), f"phase = {phase}: test_input_examples_text = {test_input_examples[0].text} != {input_examples[phase][0].text}"
            assert (
                test_input_examples[0].tags == input_examples[phase][0].tags
            ), f"phase = {phase}: test_input_examples_tags = {test_input_examples[0].tags} != {input_examples[phase][0].tags}"

        ##################################
        # 2. DataProcessor
        ##################################
        # a. get_input_examples_train
        test_input_examples, test_tag_list = data_preprocessor.get_input_examples_train(
            prune_ratio={"train": 0.5, "val": 1.0, "test": 1.0},
            dataset_name=None,
        )
        assert set(test_tag_list) == set(
            tag_list
        ), f"test_tag_list = {test_tag_list} != {tag_list}"
        for phase in ["train", "val", "test"]:
            assert (
                len(test_input_examples[phase]) == 1
            ), f"ERROR! len(test_input_examples[{phase}]) = {len(test_input_examples[phase])} should be 1."
            assert (
                test_input_examples[phase][0].text == input_examples[phase][0].text
            ), f"phase = {phase}: test_input_examples.text = {test_input_examples[phase].text} != {input_examples[phase][0].text}"
            assert (
                test_input_examples[phase][0].tags == input_examples[phase][0].tags
            ), f"phase = {phase}: test_input_examples.tags = {test_input_examples[phase].tags} != {input_examples[phase][0].tags}"

        # b. get_input_examples_predict
        test_sentences = [
            elem.text for v in test_input_examples.values() for elem in v
        ]  # retrieve example sentences
        test_input_examples_predict = data_preprocessor.get_input_examples_predict(
            test_sentences
        )["predict"]
        assert len(test_input_examples_predict) == len(
            input_examples["predict"]
        ), f"len(test_input_examples_predict) = {len(test_input_examples_predict)} != {len(input_examples['predict'])}"
        for (test_input_example_predict, true_input_example_predict) in zip(
            test_input_examples_predict, input_examples["predict"]
        ):
            assert (
                test_input_example_predict.text == true_input_example_predict.text
            ), f"test_input_example_predict.text = {test_input_example_predict.text} != {true_input_example_predict.text}"
            true_input_example_predict_tags = " ".join(
                "O" for _ in range(len(true_input_example_predict.text.split()))
            )
            assert (
                test_input_example_predict.tags == true_input_example_predict_tags
            ), f"test_input_example_predict.tags = {test_input_example_predict.tags} != {true_input_example_predict_tags}"

        # c. to_dataloader
        dataloader = data_preprocessor.to_dataloader(
            input_examples, tag_list, batch_size=1
        )
        for key in ["train", "val", "test", "predict"]:
            assert (
                key in dataloader.keys()
            ), f"key = {key} not in dataloader.keys() = {dataloader.keys()}"
            # TODO: further testing


class TestInputExamplesToTensorsAndBertDataset:
    @pytest.mark.parametrize(
        "texts, "
        "labels, "
        "tag_tuple, "
        "max_seq_length, "
        "true_input_ids, "
        "true_attention_mask, "
        "true_token_type_ids, "
        "true_tag_ids, "
        "true_input_tokens",
        [
            # 1. single example: no truncation
            (
                ["arbetsförmedlingen ai-center finns i stockholm"],
                ["ORG ORG O O LOC"],
                ("O", "ORG", "LOC"),
                12,
                torch.tensor(
                    [[101, 7093, 2842, 8126, 1011, 5410, 1121, 1045, 1305, 102, 0, 0]]
                ),
                torch.tensor([[1] * 10 + [0] * 2]),
                torch.tensor([[0] * 12]),
                torch.tensor(
                    [[-100, 1, -100, 1, -100, -100, 0, 0, 2, -100, -100, -100]]
                ),
                [
                    [
                        "[CLS]",
                        "arbetsförmedl",
                        "##ingen",
                        "ai",
                        "-",
                        "center",
                        "finns",
                        "i",
                        "stockholm",
                        "[SEP]",
                    ]
                    + ["[PAD]"] * 2
                ],
            ),
            # 2. single example: no truncation, [NEWLINE]
            (
                ["arbetsförmedlingen ai-center [NEWLINE] finns i stockholm"],
                ["ORG ORG O O O LOC"],
                ("O", "ORG", "LOC"),
                12,
                torch.tensor(
                    [
                        [
                            101,
                            7093,
                            2842,
                            8126,
                            1011,
                            5410,
                            30523,
                            1121,
                            1045,
                            1305,
                            102,
                            0,
                        ]
                    ]
                ),
                torch.tensor([[1] * 11 + [0] * 1]),
                torch.tensor([[0] * 12]),
                torch.tensor([[-100, 1, -100, 1, -100, -100, 0, 0, 0, 2, -100, -100]]),
                [
                    [
                        "[CLS]",
                        "arbetsförmedl",
                        "##ingen",
                        "ai",
                        "-",
                        "center",
                        "[NEWLINE]",
                        "finns",
                        "i",
                        "stockholm",
                        "[SEP]",
                    ]
                    + ["[PAD]"] * 1
                ],
            ),
            # 3. single example: truncation
            (
                ["arbetsförmedlingen ai-center finns i stockholm"],
                ["ORG ORG O O LOC"],
                ("O", "ORG", "LOC"),
                4,
                torch.tensor(
                    [
                        [101, 7093, 2842, 102],
                        [101, 8126, 1011, 102],
                        [101, 5410, 1121, 102],
                        [101, 1045, 1305, 102],
                    ]
                ),
                torch.tensor(
                    [
                        [1] * 4,
                        [1] * 4,
                        [1] * 4,
                        [1] * 4,
                    ]
                ),
                torch.tensor(
                    [
                        [0] * 4,
                        [0] * 4,
                        [0] * 4,
                        [0] * 4,
                    ]
                ),
                torch.tensor(
                    [
                        [-100, 1, -100, -100],
                        [-100, 1, -100, -100],
                        [-100, -100, 0, -100],
                        [-100, 0, 2, -100],
                    ]
                ),
                [
                    ["[CLS]", "arbetsförmedl", "##ingen", "[SEP]"],
                    ["[CLS]", "ai", "-", "[SEP]"],
                    ["[CLS]", "center", "finns", "[SEP]"],
                    ["[CLS]", "i", "stockholm", "[SEP]"],
                ],
            ),
            # 4. two examples: truncation
            (
                ["arbetsförmedlingen ai-center", "finns i stockholm"],
                ["ORG ORG", "O O LOC"],
                ("O", "ORG", "LOC"),
                4,
                torch.tensor(
                    [
                        [101, 7093, 2842, 102],
                        [101, 8126, 1011, 102],
                        [101, 5410, 102, 0],
                        [101, 1121, 1045, 102],
                        [101, 1305, 102, 0],
                    ]
                ),
                torch.tensor(
                    [
                        [1] * 4,
                        [1] * 4,
                        [1] * 3 + [0],
                        [1] * 4,
                        [1] * 3 + [0],
                    ]
                ),
                torch.tensor(
                    [
                        [0] * 4,
                        [0] * 4,
                        [0] * 4,
                        [0] * 4,
                        [0] * 4,
                    ]
                ),
                torch.tensor(
                    [
                        [-100, 1, -100, -100],
                        [-100, 1, -100, -100],
                        [-100, -100, -100, -100],
                        [-100, 0, 0, -100],
                        [-100, 2, -100, -100],
                    ]
                ),
                [
                    ["[CLS]", "arbetsförmedl", "##ingen", "[SEP]"],
                    ["[CLS]", "ai", "-", "[SEP]"],
                    ["[CLS]", "center", "[SEP]", "[PAD]"],
                    ["[CLS]", "finns", "i", "[SEP]"],
                    ["[CLS]", "stockholm", "[SEP]", "[PAD]"],
                ],
            ),
        ],
    )
    def tests(
        self,
        texts: List[str],
        labels: List[str],
        tag_tuple: List[str],
        max_seq_length: int,
        true_input_ids: torch.Tensor,
        true_attention_mask: torch.Tensor,
        true_token_type_ids: torch.Tensor,
        true_tag_ids: torch.Tensor,
        true_input_tokens: torch.Tensor,
    ) -> None:

        ##################################
        # 1. InputExamplesToTensors
        ##################################
        input_examples = [
            InputExample(
                guid="",
                text=text,
                tags=label,
            )
            for text, label in zip(texts, labels)
        ]
        input_examples_to_tensors = InputExamplesToTensors(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            tag_tuple=tuple(tag_tuple),
            default_logger=PseudoDefaultLogger(),
        )
        encodings = input_examples_to_tensors(input_examples, predict=False)
        input_tokens = [
            tokenizer.convert_ids_to_tokens(input_ids_single)
            for input_ids_single in encodings["input_ids"]
        ]

        for (string, true) in zip(
            EncodingsKeys,
            [true_input_ids, true_attention_mask, true_token_type_ids, true_tag_ids],
        ):
            assert torch.all(
                torch.eq(encodings[string], true)
            ), f"{string} = {encodings[string]} != {true}"

        for (string, true, _test) in zip(
            ["input_tokens"],
            [true_input_tokens],
            [input_tokens],
        ):
            assert _test == true, f"{string} = {_test} != {true}"

        ##################################
        # 2. BertDataset
        ##################################
        data = BertDataset(
            encodings=encodings
        )  # data[j] = 4 torch tensors corresponding to EncodingKeys
        assert len(data) >= len(
            texts
        ), f"len(data) = {len(data)} < {len(texts)} = len(texts)"
        for i, (string, true) in enumerate(
            zip(
                ["input_ids", "attention_mask", "token_type_ids", "tag_ids"],
                [
                    true_input_ids,
                    true_attention_mask,
                    true_token_type_ids,
                    true_tag_ids,
                ],
            )
        ):
            for j in range(len(true)):
                assert torch.all(
                    torch.eq(data[j][i], true[j])
                ), f"{string} = {data[j][i]} != {true[j]}"


class TestMisc:

    ####################################################################################################################
    @pytest.mark.parametrize(
        "tag_list, " "returned_tag_list",
        [
            (
                ["O", "PER", "ORG", "MISC"],
                ["O", "PER", "ORG", "MISC"],
            ),
            (
                ["O", "B-PER", "B-ORG", "B-MISC"],
                ["O", "B-PER", "B-ORG", "B-MISC", "I-PER", "I-ORG", "I-MISC"],
            ),
        ],
    )
    def tests(
        self,
        tag_list: List[str],
        returned_tag_list: List[str],
    ) -> None:
        test_returned_tag_list = (
            data_preprocessor._ensure_completeness_in_case_of_bio_tags(
                tag_list=tag_list
            )
        )
        assert (
            test_returned_tag_list == returned_tag_list
        ), f"test_returned_tag_list = {test_returned_tag_list} != {returned_tag_list}"


if __name__ == "__main__":
    TestCsvReaderAndDataProcessor().tests()
    TestInputExamplesToTensorsAndBertDataset().tests()
    TestMisc().tests()



import pytest
from typing import List
from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics


class TestLoggedMetrics:

    logged_metrics = LoggedMetrics()

    @pytest.mark.parametrize(
        "tag_group, " "phase_group, " "micro_macro_group, " "exclude, " "metrics",
        [
            (
                ["all"],
                ["train"],
                None,
                None,
                ["loss", "acc"],
            ),
            (
                ["all"],
                ["train"],
                None,
                ["acc"],
                ["loss"],
            ),
            (
                ["ind"],
                ["val"],
                None,
                None,
                ["f1", "precision", "recall"],
            ),
            (
                ["ind"],
                ["val"],
                ["micro"],
                None,
                ["f1", "precision", "recall"],
            ),
            (
                ["ind"],
                ["val"],
                ["macro"],
                None,
                [],
            ),
            (
                None,
                None,
                None,
                None,
                ["loss", "precision", "acc", "recall", "f1"],
            ),
        ],
    )
    def test_get_metrics(
        self,
        tag_group: List[str],
        phase_group: List[str],
        micro_macro_group: List[str],
        exclude: List[str],
        metrics: List[str],
    ):

        test_metrics = self.logged_metrics.get_metrics(
            tag_group=tag_group,
            phase_group=phase_group,
            micro_macro_group=micro_macro_group,
            exclude=exclude,
        )
        assert set(test_metrics) == set(metrics), f"ERROR!"

    @pytest.mark.parametrize(
        "metrics_flat_list",
        [
            (
                [
                    "all_loss",
                    "all_acc",
                    #
                    "all_precision_micro",
                    "all_precision_macro",
                    "all_recall_micro",
                    "all_recall_macro",
                    "all_f1_micro",
                    "all_f1_macro",
                    #
                    "fil_precision_micro",
                    "fil_precision_macro",
                    "fil_recall_micro",
                    "fil_recall_macro",
                    "fil_f1_micro",
                    "fil_f1_macro",
                    #
                    "chk_precision_micro",
                    "chk_recall_micro",
                    "chk_f1_micro",
                    #
                    "ind_precision_micro",
                    "ind_recall_micro",
                    "ind_f1_micro",
                ]
            )
        ],
    )
    def test_as_flat_list(self, metrics_flat_list: List[str]):
        test_metrics_flat_list = self.logged_metrics.as_flat_list()
        assert set(test_metrics_flat_list) == set(metrics_flat_list), f"ERROR! 2"



import os
import subprocess
from os.path import join

from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter
from nerblackbox.modules.utils.env_variable import env_variable


class SICFormatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "sic"
        ner_tag_list = [
            "person",
            "animal",
            "myth",
            "place",
            "inst",
            "product",
            "work",
            "event",
            "other",
        ]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool):
        """
        I: get data
        ----------
        :param verbose: [bool]
        :return: -
        """
        bash_cmds = [
            f"mkdir {env_variable('DIR_DATASETS')}/_sic",
            f"curl -o {env_variable('DIR_DATASETS')}/_sic/sic.zip "
            "https://www.ling.su.se/polopoly_fs/1.99145.1380811903\!/menu/standard/file/sic.zip",
            f"cd {env_variable('DIR_DATASETS')}/_sic && unzip -o sic.zip",
            f"mkdir {env_variable('DIR_DATASETS')}/sic/raw_data",
            f"mv {env_variable('DIR_DATASETS')}/_sic/sic/annotated/* {env_variable('DIR_DATASETS')}/sic/raw_data",
            f"rm -r {env_variable('DIR_DATASETS')}/_sic",
            f"cat {env_variable('DIR_DATASETS')}/sic/raw_data/*.conll "
            f"> {env_variable('DIR_DATASETS')}/sic/sic-train.conll",
        ]

        for bash_cmd in bash_cmds:
            if verbose:
                print(bash_cmd)

            try:
                subprocess.run(bash_cmd, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                print(e)

    def create_ner_tag_mapping(self):
        """
        II: customize ner_training tag mapping if wanted
        ------------------------------------------------
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(self):
        """
        III: format data
        ----------------
        :return: -
        """
        for phase in ["train"]:
            rows = self._read_original_file(phase)
            self._write_formatted_csv(phase, rows)

    def resplit_data(self, val_fraction: float):
        """
        IV: resplit data
        ----------------
        :param val_fraction: [float]
        :return: -
        """
        # train -> train, val, test
        df_train_val_test = self._read_formatted_csvs(["train"])
        df_train_val, df_test = self._split_off_validation_set(
            df_train_val_test, val_fraction
        )
        df_train, df_val = self._split_off_validation_set(df_train_val, val_fraction)
        self._write_final_csv("train", df_train)
        self._write_final_csv("val", df_val)
        self._write_final_csv("test", df_test)

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    def _read_original_file(self, phase):
        """
        III: format data
        ---------------------------------------------
        :param phase:   [str] 'train' or 'test'
        :return: _rows: [list] of [list] of [str], e.g. [[], ['Inger', 'PER'], ['säger', '0'], ..]
        """
        file_name = {
            "train": "sic-train.conll",
        }
        file_path_original = join(self.dataset_path, file_name[phase])

        _rows = list()
        if os.path.isfile(file_path_original):
            with open(file_path_original) as f:
                for i, row in enumerate(f.readlines()):
                    _rows.append(row.strip().split())
            print(f"\n> read {file_path_original}")
        else:
            raise Exception(f"> original file {file_path_original} could not be found.")

        _rows = [
            [row[1], self.transform_tags(row[-3], row[-2])]
            if len(row) > 0
            and row[-3]
            not in ["_", "U"]  # filter out "_-_" and "_-person" (bugs in SIC dataset)
            else list()
            for row in _rows
        ]

        return _rows

    @staticmethod
    def transform_tags(bio, tag):
        """
        :param bio: [str] 'O', 'B', 'I'
        :param tag: [str] '_', 'person', ..
        :return: transformed tag: [str], e.g. 'O', 'B-person', ..
        """
        if bio == "O":
            return "O"
        else:
            return f"{bio}-{tag}"



import torch
import mlflow
import os
from os.path import join
from pytorch_lightning import Trainer
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.callbacks import EarlyStopping

from nerblackbox.modules.ner_training.ner_model_train import (
    NerModelTrain,
)
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
from nerblackbox.modules.utils.util_functions import unify_parameters
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.utils.util_functions import (
    get_package_version,
    checkpoint2epoch,
    epoch2checkpoint,
)


def main(params, hparams, log_dirs, experiment: bool):
    """
    :param params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param hparams:    [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param log_dirs:   [argparse.Namespace] attr: mlflow, tensorboard
    :param experiment: [bool] whether run is part of an experiment w/ multiple runs
    :return: -
    """
    default_logger = DefaultLogger(
        __file__, log_file=log_dirs.log_file, level=params.logging_level
    )  # python logging
    default_logger.clear()

    print_run_information(params, hparams, default_logger)
    lightning_hparams = unify_parameters(params, hparams, log_dirs, experiment)

    tb_logger = logging_start(params, log_dirs)
    with mlflow.start_run(run_name=params.run_name_nr, nested=experiment):

        model = NerModelTrain(lightning_hparams)
        callbacks = get_callbacks(params, hparams, log_dirs)

        trainer = Trainer(
            max_epochs=hparams.max_epochs,
            gpus=torch.cuda.device_count() if params.device.type == "cuda" else None,
            precision=16 if (params.fp16 and params.device.type == "cuda") else 32,
            amp_level="O1",
            logger=tb_logger,
            checkpoint_callback=callbacks["checkpoint"],
            early_stop_callback=callbacks["early_stop"],
        )
        trainer.fit(model)
        trainer.test()

        callback_info = get_callback_info(callbacks, params, hparams)

        # use best checkpoint
        model_best = NerModelTrain.load_from_checkpoint(
            checkpoint_path=callback_info["checkpoint_best"]
        )
        best_trainer = Trainer(
            gpus=torch.cuda.device_count() if params.device.type == "cuda" else None,
            precision=16 if (params.fp16 and params.device.type == "cuda") else 32,
            logger=tb_logger,
        )
        best_trainer.test(model_best)

        # logging end
        logging_end(
            tb_logger, callback_info, hparams, model, model_best, default_logger
        )

        # remove checkpoint
        if params.checkpoints is False:
            remove_checkpoint(callback_info["checkpoint_best"], default_logger)


########################################################################################################################
# HELPER FUNCTIONS #####################################################################################################
########################################################################################################################
def print_run_information(_params, _hparams, _logger):
    """
    :param _params:   [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:  [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, prune_ratio_*, lr_*
    :param _logger:   [DefaultLogger]
    :return: -
    """
    _logger.log_info(f">>> NER BLACK BOX VERSION: {get_package_version()}")
    _logger.log_info("- PARAMS -----------------------------------------")
    _logger.log_info(f"> experiment_name: {_params.experiment_name}")
    _logger.log_info(f"> run_name_nr:     {_params.run_name_nr}")
    _logger.log_info("..")
    _logger.log_info(f"> available GPUs: {torch.cuda.device_count()}")
    _logger.log_info(f"> device:         {_params.device}")
    _logger.log_info(f"> fp16:           {_params.fp16}")
    _logger.log_info("..")
    _logger.log_info(f"> dataset_name:          {_params.dataset_name}")
    _logger.log_info(f"> dataset_tags:          {_params.dataset_tags}")
    _logger.log_info(f"> prune_ratio_train:     {_params.prune_ratio_train}")
    _logger.log_info(f"> prune_ratio_val:       {_params.prune_ratio_val}")
    _logger.log_info(f"> prune_ratio_test:      {_params.prune_ratio_test}")
    _logger.log_info("..")
    _logger.log_info(f"> pretrained_model_name: {_params.pretrained_model_name}")
    _logger.log_info(f"> uncased:               {_params.uncased}")
    _logger.log_info("..")
    _logger.log_info(f"> checkpoints:           {_params.checkpoints}")
    _logger.log_info(f"> logging_level:         {_params.logging_level}")
    _logger.log_info(f"> multiple_runs:         {_params.multiple_runs}")
    _logger.log_info("")
    _logger.log_info("- HPARAMS ----------------------------------------")
    _logger.log_info(f"> batch_size:       {_hparams.batch_size}")
    _logger.log_info(f"> max_seq_length:   {_hparams.max_seq_length}")
    _logger.log_info(f"> max_epochs:       {_hparams.max_epochs}")
    _logger.log_info(f"> monitor:          {_hparams.monitor}")
    _logger.log_info(f"> min_delta:        {_hparams.min_delta}")
    _logger.log_info(f"> patience:         {_hparams.patience}")
    _logger.log_info(f"> mode:             {_hparams.mode}")
    _logger.log_info(f"> lr_max:           {_hparams.lr_max}")
    _logger.log_info(f"> lr_warmup_epochs: {_hparams.lr_warmup_epochs}")
    _logger.log_info(f"> lr_schedule:      {_hparams.lr_schedule}")
    _logger.log_info(f"> lr_num_cycles:    {_hparams.lr_num_cycles}")
    _logger.log_info("")


def _get_model_checkpoint_directory(_params):
    """
    :param _params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :return: model_checkpoint_directory [str]
    """
    return join(env_variable("DIR_CHECKPOINTS"), _params.experiment_run_name_nr)


def get_callbacks(_params, _hparams, _log_dirs):
    """
    :param _params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:    [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param _log_dirs:   [argparse.Namespace] attr: mlflow, tensorboard
    :return: _callbacks: [dict] w/ keys 'checkpoint', 'early_stop' & values = [pytorch lightning callback]
    """
    early_stopping_params = {
        k: vars(_hparams)[k] for k in ["monitor", "min_delta", "patience", "mode"]
    }
    model_checkpoint_filepath = join(
        _get_model_checkpoint_directory(_params), "{epoch}"
    )

    _callbacks = {
        "checkpoint": ModelCheckpoint(
            filepath=model_checkpoint_filepath, verbose=True
        ),  # if _params.checkpoints else None,
        "early_stop": EarlyStopping(**early_stopping_params, verbose=True),
    }
    return _callbacks


def get_callback_info(_callbacks, _params, _hparams):
    """
    :param _callbacks: [dict] w/ keys 'checkpoint', 'early_stop' & values = [pytorch lightning callback]
    :param _params:    [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _hparams:   [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :return: _callback_info: [dict] w/ keys 'epoch_best', 'epoch_stopped', 'checkpoint_best'
    """
    checkpoint_best = list(_callbacks["checkpoint"].best_k_models.keys())[0]
    callback_info = dict()
    callback_info["epoch_best"] = checkpoint2epoch(checkpoint_best)
    callback_info["epoch_stopped"] = (
        _callbacks["early_stop"].stopped_epoch
        if _callbacks["early_stop"].stopped_epoch
        else _hparams.max_epochs - 1
    )
    callback_info["checkpoint_best"] = join(
        _get_model_checkpoint_directory(_params),
        epoch2checkpoint(callback_info["epoch_best"]),
    )

    return callback_info


def logging_start(_params, _log_dirs):
    """
    :param _params:      [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
    :param _log_dirs:    [argparse.Namespace] attr: mlflow, tensorboard
    :return: _tb_logger: [pytorch lightning TensorBoardLogger]
    """
    # mlflow.tracking.set_tracking_uri(_log_dirs.mlflow)                # mlflow
    # mlflow.set_experiment(_params.experiment_name)                    # mlflow
    _tb_logger = TensorBoardLogger(
        save_dir=_log_dirs.tensorboard,  # tensorboard
        name=_params.experiment_run_name_nr,
    )
    return _tb_logger


def logging_end(
    _tb_logger, _callback_info, _hparams, _model_stopped, _model_best, _logger
):
    """
    :param _tb_logger:     [pytorch lightning TensorBoardLogger]
    :param _callback_info: [dict] w/ keys 'epoch_best', 'epoch_stopped', 'checkpoint_best'
    :param _hparams:       [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, lr_*
    :param _model_stopped: [NerModelTrain] epoch_stopped
    :param _model_best:    [NerModelTrain] epoch_best
    :param _logger:        [DefaultLogger]
    :return: -
    """
    epoch_best = _callback_info["epoch_best"]
    epoch_stopped = _callback_info["epoch_stopped"]

    _logger.log_info(f"epoch_best: {epoch_best}")
    _logger.log_info(f"epoch_stopped: {epoch_stopped}")

    _model_stopped.mlflow_client.log_metric("epoch_best", epoch_best)
    _model_stopped.mlflow_client.log_metric("epoch_stopped", epoch_stopped)
    for metric in ("all_f1_micro", "fil_f1_micro", "chk_f1_micro"):
        _model_stopped.mlflow_client.log_metric(
            f"epoch_stopped_val_{metric}",
            _model_stopped.epoch_metrics["val"][epoch_stopped][metric],
        )
        _model_stopped.mlflow_client.log_metric(
            f"epoch_stopped_test_{metric}",
            _model_stopped.epoch_metrics["test"][epoch_stopped][metric],
        )
        _model_stopped.mlflow_client.log_metric(
            f"epoch_best_val_{metric}",
            _model_stopped.epoch_metrics["val"][epoch_best][metric],
        )
        _model_stopped.mlflow_client.log_metric(
            f"epoch_best_test_{metric}", _model_best.epoch_metrics["test"][0][metric]
        )

    _model_stopped.mlflow_client.finish_artifact_logger()  # mlflow
    _tb_logger_stopped_epoch(
        _tb_logger, _hparams, epoch_best, epoch_stopped, _model_stopped, _model_best
    )  # tb


def _tb_logger_stopped_epoch(
    _tb_logger,
    _hparams,
    _epoch_best,
    _epoch_stopped,
    _model_stopped,
    _model_best,
    metrics=("all_f1_micro", "fil_f1_micro", "chk_f1_micro"),
):
    """
    log hparams and metrics for stopped epoch
    -----------------------------------------
    :param _tb_logger:      [pytorch lightning TensorboardLogger]
    :param _hparams:        [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, prune_ratio_*, lr_*
    :param _epoch_best:     [int]
    :param _epoch_stopped:  [int]
    :param _model_stopped:  [NerModelTrain] epoch_stopped
    :param _model_best:     [NerModelTrain] epoch_best
    :param metrics:         [tuple] of metrics to be logged in hparams section
    :return:
    """
    hparams_dict = dict()

    # epoch_best & epoch_stopped
    hparams_dict["hparam/train/epoch_best"] = _epoch_best
    hparams_dict["hparam/train/epoch_stopped"] = _epoch_stopped

    # val/test
    hparams_stopped_val = {
        f'hparam/val/epoch_stopped_{_epoch_stopped}/{metric.replace("+", "P")}': _model_stopped.epoch_metrics[
            "val"
        ][
            _epoch_stopped
        ][
            metric
        ]
        for metric in metrics
    }
    hparams_stopped_test = {
        f'hparam/test/epoch_stopped_{_epoch_stopped}/{metric.replace("+", "P")}': _model_stopped.epoch_metrics[
            "test"
        ][
            _epoch_stopped
        ][
            metric
        ]
        for metric in metrics
    }

    hparams_best_val = {
        f'hparam/val/epoch_best_{_epoch_best}/{metric.replace("+", "P")}': _model_stopped.epoch_metrics[
            "val"
        ][
            _epoch_best
        ][
            metric
        ]
        for metric in metrics
    }

    hparams_best_test = {
        f'hparam/test/epoch_best_{_epoch_best}/{metric.replace("+", "P")}': _model_best.epoch_metrics[
            "test"
        ][
            0
        ][
            metric
        ]
        for metric in metrics
    }

    hparams_dict.update(hparams_stopped_val)
    hparams_dict.update(hparams_stopped_test)
    hparams_dict.update(hparams_best_val)
    hparams_dict.update(hparams_best_test)

    """
    NOTE: 
    The following command (= that adds hyperparameters (_hparams) & most important metrics (hparams_dict) to tb logging)
    logs properly to tensorboard's SCALAR, but not to HPARAMS.
    Apparently it is incompatible with pytorch-lightning's default tensorboard logging, 
    see https://github.com/PyTorchLightning/pytorch-lightning/issues/1225
    So technically speaking there are 2 logs for each run, pytorch-lightning's default & the one below. 
    """
    # log
    _tb_logger.experiment.add_hparams(
        vars(_hparams),
        hparams_dict,
    )


def remove_checkpoint(_checkpoint_path, _default_logger):
    """
    remove checkpoint stored at _checkpoint_path
    --------------------------------------------
    :param _checkpoint_path: [str], e.g. '[..]/data/results/checkpoints/exp_default/run1/epoch=0.ckpt'
    :param _default_logger:  [DefaultLogger]
    :return: -
    """
    os.remove(_checkpoint_path)
    _default_logger.log_info(f"> checkpoint {_checkpoint_path} removed")



import os
import subprocess
import json
import pandas as pd
from abc import ABC, abstractmethod

from os.path import join

from nerblackbox.modules.utils.util_functions import get_dataset_path
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.datasets.plots import Plots
from nerblackbox.modules.datasets.formatter.util_functions import get_ner_tag_mapping
from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger


class BaseFormatter(ABC):
    def __init__(self, ner_dataset, ner_tag_list):
        """
        :param ner_dataset:  [str] 'swedish_ner_corpus' or 'suc'
        :param ner_tag_list: [list] of [str], e.g. ['PER', 'LOC', ..]
        """
        self.ner_dataset = ner_dataset
        self.ner_tag_list = ner_tag_list
        self.dataset_path = get_dataset_path(ner_dataset)
        self.stats_aggregated = None
        self.num_tokens = None
        self.num_sentences = None

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    @abstractmethod
    def get_data(self, verbose: bool):
        """
        I: get data
        -----------
        :param verbose: [bool]
        :return: -
        """
        pass

    @abstractmethod
    def create_ner_tag_mapping(self):
        """
        II: customize ner_training tag mapping if wanted
        -------------------------------------
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        pass

    @abstractmethod
    def format_data(self):
        """
        III: format data
        ----------------
        :return: -
        """
        pass

    @abstractmethod
    def resplit_data(self, val_fraction: float):
        """
        IV: resplit data
        ----------------
        :param val_fraction: [float]
        :return: -
        """
        pass

    ####################################################################################################################
    # BASE METHODS
    ####################################################################################################################
    def create_directory(self):
        """
        0: create directory for dataset
        -------------------------------
        :return: -
        """
        directory_path = (
            f'{env_variable("DIR_DATASETS")}/{self.ner_dataset}/analyze_data'
        )
        os.makedirs(directory_path, exist_ok=True)

        bash_cmd = (
            f'echo "*" > {env_variable("DIR_DATASETS")}/{self.ner_dataset}/.gitignore'
        )
        try:
            subprocess.run(bash_cmd, shell=True, check=True)
        except subprocess.CalledProcessError as e:
            print(e)

    def create_ner_tag_mapping_json(self, modify: bool):
        """
        II: create customized ner_training tag mapping to map tags in original data to tags in formatted data
        ----------------------------------------------------------------------------------------------
        :param modify:      [bool], if True: modify tags as specified in method modify_ner_tag_mapping()
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        if modify:
            ner_tag_mapping = self.create_ner_tag_mapping()
        else:
            ner_tag_mapping = dict()

        json_path = (
            f'{env_variable("DIR_DATASETS")}/{self.ner_dataset}/ner_tag_mapping.json'
        )
        with open(json_path, "w") as f:
            json.dump(ner_tag_mapping, f)

        print(f"> dumped the following dict to {json_path}:")
        print(ner_tag_mapping)

    ####################################################################################################################
    # HELPER: WRITE FORMATTED
    ####################################################################################################################
    def _write_formatted_csv(self, phase, rows):
        """
        III: format data
        ----------------------------------------------
        :param phase:         [str] 'train' or 'test'
        :param rows:          [list] of [list] of [str], e.g. [['Inger', 'PER'], ['säger', '0'], ..]
        :return: -
        """
        file_path = join(self.dataset_path, f"{phase}_formatted.csv")

        # ner tag mapping
        ner_tag_mapping = get_ner_tag_mapping(
            path=join(self.dataset_path, "ner_tag_mapping.json")
        )

        # processing
        data = list()
        tags = list()
        sentence = list()
        for row in rows:
            if len(row) == 2:
                sentence.append(row[0])
                tags.append(
                    ner_tag_mapping(row[1]) if row[1] != "0" else "O"
                )  # replace zeros by capital O (!)
            else:
                if len(row) != 0:
                    print(
                        f"ATTENTION!! row with length = {len(row)} found (should be 0 or 2): {row}"
                    )
                if len(tags) and len(sentence):
                    data.append([" ".join(tags), " ".join(sentence)])
                    tags = list()
                    sentence = list()

        df = pd.DataFrame(data)
        df.to_csv(file_path, sep="\t", header=None, index=None)
        print(
            f"> phase = {phase}: wrote {len(rows)} words in {len(df)} sentences to {file_path}"
        )

    ####################################################################################################################
    # HELPER: READ FORMATTED
    ####################################################################################################################
    def _read_formatted_csvs(self, phases):
        """
        IV: resplit data
        ----------------
        :param phases: [list] of [str] to read formatted csvs from, e.g. ['val', 'test']
        :return: [pd DataFrame]
        """
        df_phases = [self._read_formatted_csv(phase) for phase in phases]
        return pd.concat(df_phases, ignore_index=True)

    def _read_formatted_csv(self, phase):
        """
        IV: resplit data
        ----------------
        :param phase: [str] csvs to read formatted df from, e.g. 'val'
        :return: df: [pd DataFrame]
        """
        formatted_file_path = join(self.dataset_path, f"{phase}_formatted.csv")
        try:
            df = pd.read_csv(formatted_file_path, sep="\t", header=None)
        except pd.errors.EmptyDataError:
            df = pd.DataFrame()
        return df

    ####################################################################################################################
    # HELPER: WRITE FINAL
    ####################################################################################################################
    @staticmethod
    def _split_off_validation_set(_df_original, _val_fraction):
        """
        IV: resplit data
        ----------------
        :param _df_original:    [pd DataFrame]
        :param _val_fraction:   [float] between 0 and 1
        :return: _df_new:       [pd DataFrame]
        :return: _df_val:       [pd DataFrame]
        """
        split_index = int(len(_df_original) * (1 - _val_fraction))
        _df_new = _df_original.iloc[:split_index]
        _df_val = _df_original.iloc[split_index:]
        return _df_new, _df_val

    def _write_final_csv(self, phase, df):
        """
        IV: resplit data
        ----------------
        :param phase: [str], 'train', 'val' or 'test'
        :param df: [pd DataFrame]
        :return:
        """
        file_path = join(self.dataset_path, f"{phase}.csv")
        df.to_csv(file_path, sep="\t", index=False, header=None)

    ####################################################################################################################
    def read_formatted_csv(self, phase):
        """
        V: read formatted csv files
        ----------------------------------------------
        :param phase:         [str] 'train' or 'test'
        :return: num_sentences:    [int]
                 stats_aggregated: [pandas Series] with indices = tags, values = number of occurrences
        """
        file_path = join(self.dataset_path, f"{phase}.csv")

        columns = ["O"] + self.ner_tag_list

        try:
            df = pd.read_csv(file_path, sep="\t")
        except pd.io.common.EmptyDataError:
            df = None

        stats = pd.DataFrame([], columns=columns)

        if df is not None:
            tags = (
                df.iloc[:, 0]
                .apply(lambda x: x.split())
                .apply(lambda x: [elem.split("-")[-1] for elem in x])
            )

            for column in columns:
                stats[column] = tags.apply(
                    lambda x: len([elem for elem in x if elem == column])
                )

            assert len(df) == len(stats)

        num_sentences = len(stats)
        stats_aggregated = stats.sum().to_frame().astype(int)
        stats_aggregated.columns = ["tags"]

        return num_sentences, stats_aggregated

    @staticmethod
    def get_tokens(df):
        return df.loc[:, "tags"].sum()

    def analyze_data(self):
        """
        V: analyze data
        ----------------
        :created attr: stats_aggregated: [dict] w/ keys = 'total', 'train', 'val', 'test' & values = [df]
        :return: -
        """
        log_file = join(self.dataset_path, "analyze_data", f"{self.ner_dataset}.log")
        default_logger = DefaultLogger(
            __file__, log_file=log_file, level="info", mode="w"
        )

        self.num_tokens = {"total": 0}
        self.num_sentences = {"total": 0}
        self.stats_aggregated = {"total": None}

        phases = ["train", "val", "test"]
        phases_all = ["total"] + phases

        for phase in phases:
            (
                self.num_sentences[phase],
                _stats_aggregated_phase,
            ) = self.read_formatted_csv(phase)
            self.num_sentences["total"] += self.num_sentences[phase]
            if self.stats_aggregated["total"] is None:
                self.stats_aggregated["total"] = _stats_aggregated_phase
            else:
                self.stats_aggregated["total"] = (
                    self.stats_aggregated["total"] + _stats_aggregated_phase
                )

            self.stats_aggregated[phase] = self._stats_aggregated_add_columns(
                _stats_aggregated_phase, self.num_sentences[phase]
            )

        num_sentences_total = self.num_sentences["total"]
        self.stats_aggregated["total"] = self._stats_aggregated_add_columns(
            self.stats_aggregated["total"], self.num_sentences["total"]
        )
        self.num_tokens = {
            phase: self.get_tokens(self.stats_aggregated[phase]) for phase in phases_all
        }
        num_tokens_total = self.num_tokens["total"]

        # print/log
        for phase in phases:
            default_logger.log_info("")
            default_logger.log_info(f">>> {phase} <<<<")
            default_logger.log_info(
                f"num_sentences = {self.num_sentences[phase]} "
                f"({100*self.num_sentences[phase]/num_sentences_total:.2f}% of total = {num_sentences_total})"
            )
            default_logger.log_info(
                f"num_tokens = {self.num_tokens[phase]} "
                f"({100*self.num_tokens[phase]/num_tokens_total:.2f}% of total = {num_tokens_total})"
            )
            default_logger.log_info(self.stats_aggregated[phase])

        default_logger.log_info("")
        default_logger.log_info(f"num_sentences = {self.num_sentences}")
        default_logger.log_info(f"num_tokens = {self.num_tokens}")
        default_logger.log_info(f">>> total <<<<")
        default_logger.log_info(self.stats_aggregated["total"])

    def plot_data(self):
        fig_path = join(self.dataset_path, "analyze_data", f"{self.ner_dataset}.png")
        Plots(self.stats_aggregated, self.num_sentences).plot(fig_path=fig_path)

    @staticmethod
    def _stats_aggregated_add_columns(df, number_of_sentences):
        """
        V: analyze data
        ----------------
        :param df: ..
        :param number_of_sentences: ..
        :return: ..
        """
        df["tags/sentence"] = df["tags"] / float(number_of_sentences)
        df["tags/sentence"] = df["tags/sentence"].apply(lambda x: "{:.2f}".format(x))

        # relative tags w/ 0
        number_of_occurrences = df["tags"].sum()
        df["tags relative w/ 0"] = df["tags"] / number_of_occurrences
        df["tags relative w/ 0"] = df["tags relative w/ 0"].apply(
            lambda x: "{:.2f}".format(x)
        )

        # relative tags w/o 0
        number_of_filtered_occurrences = df["tags"].sum() - df.loc["O"]["tags"]
        df["tags relative w/o 0"] = df["tags"] / number_of_filtered_occurrences
        df["tags relative w/o 0"]["O"] = 0
        df["tags relative w/o 0"] = df["tags relative w/o 0"].apply(
            lambda x: "{:.2f}".format(x)
        )

        return df



import torch
from torch.utils.data import Dataset
from typing import Tuple
from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
    Encodings,
    EncodingsKeys,
)


class BertDataset(Dataset):
    """ Bert Dataset with features created when requested """

    def __init__(self, encodings: Encodings) -> None:
        """
        Args:
            encodings: [Encodings] with
            key = input_ids,      value = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]]
            key = attention_mask, value = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]]
            key = token_type_ids, value = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]]
            key = tag_ids,        value = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]]
        """
        self.encodings = encodings

    def __len__(self) -> int:
        return self.encodings["tag_ids"].shape[0]

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, ...]:
        """
        Args:
            index: [int]

        Returns:
            input_ids:      [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]]
            attention_mask: [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]]
            token_type_ids: [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]]
            tag_ids:        [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]]
        """
        return tuple([self.encodings[key][index] for key in EncodingsKeys])



import torch
import numpy as np
from typing import List, Tuple, Dict, Any
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)
from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
    Encodings,
    EncodingsKeys,
)


class InputExamplesToTensors:
    """
    Converts List[InputExample] to Encodings w/ keys = "input_ids", "attention_mask", "token_type_ids", "tag_ids"
    """

    def __init__(
        self,
        tokenizer,
        max_seq_length: int = 128,
        tag_tuple: tuple = ("O", "PER", "ORG"),
        default_logger=None,
    ):
        """
        :param tokenizer:      [BertTokenizer] used to tokenize to Wordpieces and transform to indices
        :param max_seq_length: [int]
        :param tag_tuple:      [tuple] of [str]
        """
        self.tokenizer = tokenizer
        self.max_seq_length = max_seq_length
        self.default_logger = default_logger

        self.tag2id = {tag: i for i, tag in enumerate(tag_tuple)}
        if self.default_logger:
            self.default_logger.log_debug("> tag2id:", self.tag2id)

    def __call__(self, input_examples: List[InputExample], predict: bool) -> Encodings:
        """
        Args:
            input_examples: List[InputExample] with e.g. text = 'at arbetsförmedlingen'
                                                         tags = '0 ORG'
            predict: if True,  map special tokens to    O tag_id for prediction (will not be used)
                     if False, map special tokens to -100 tag_id for train, val, test

        Returns:
            encodings: [Encodings] with
            key = input_ids,      value = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]]
            key = attention_mask, value = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]]
            key = token_type_ids, value = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]]
            key = tag_ids,        value = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]]
        """

        # encodings_single
        encodings_single: Dict[str, List[Any]] = {key: list() for key in EncodingsKeys}
        for input_example in input_examples:
            _encodings = self._transform_input_example(input_example, predict)
            for key in encodings_single.keys():
                encodings_single[key].append(_encodings[key])

        # combine encodings_single -> encodings
        encodings = {
            key: torch.cat(encodings_single[key]) for key in encodings_single.keys()
        }

        return encodings

    def _transform_input_example(
        self, input_example: InputExample, predict: bool
    ) -> Encodings:
        """
        - transform input_example to tensors of length self.max_seq_length

        Args:
            input_example: [InputExample], with e.g. text = 'at arbetsförmedlingen'
                                                     tags = '0 ORG'
            predict: if True,  map special tokens to    O tag_id for prediction (will not be used)
                     if False, map special tokens to -100 tag_id for train, val, test

        Returns:
            encodings: [Encodings] with
            key = input_ids,      value = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]]
            key = attention_mask, value = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]]
            key = token_type_ids, value = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]]
            key = tag_ids,        value = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]]
        """
        ####################
        # A0. tokens_*, tags_*
        ####################
        tokens_split_into_words = input_example.text.split()
        tags_split_into_words = input_example.tags.split()
        assert len(tokens_split_into_words) == len(
            tags_split_into_words
        ), f"ERROR! len(tokens) = {len(tokens_split_into_words)} is different from len(tags) = {len(tags_split_into_words)}"

        encodings = self.tokenizer(
            tokens_split_into_words,
            padding="max_length",
            truncation=True,
            max_length=self.max_seq_length,
            is_split_into_words=True,
            return_offsets_mapping=True,
            stride=0,
            return_overflowing_tokens=True,
        )

        encodings["tag_ids"] = self._encode_tags(
            tags_split_into_words, encodings.offset_mapping, predict
        )
        keys = list(encodings.keys())
        for key in keys:
            if key in EncodingsKeys:
                encodings[key] = torch.tensor(encodings[key])
            else:
                encodings.pop(key)

        return encodings

    ####################################################################################################################
    # PRIVATE HELPER METHODS
    ####################################################################################################################
    def _encode_tags(
        self,
        _tags_split_into_words: List[str],
        all_offsets: List[List[Tuple[int, int]]],
        predict: bool,
    ) -> List[List[int]]:
        """
        Args:
            _tags_split_into_words: ['at arbetsförmedlingen']
            all_offsets: [chunks] with chunk = [(start_1, end_1), (start_2, end_2), ..], e.g. [[(0, 2), (2, 5), ..]
            predict: if True,  map special tokens to    O tag_id for prediction (will not be used)
                     if False, map special tokens to -100 tag_id for train, val, test

        Returns:
            all_tag_ids: [chunk_tag_ids] with chunk_tag_ids = e.g. [-100, 3, -100, 4, 5, -100, ..]
        """

        tag_ids_split_into_words: List[int] = [
            self.tag2id[tag] for tag in _tags_split_into_words
        ]

        tag_id_special = 0 if predict else -100

        index = 0
        all_tag_ids = list()
        for offsets in all_offsets:
            # create an empty array of -100
            arr_tag_ids: np.array = np.ones(len(offsets), dtype=int) * tag_id_special
            arr_offsets: np.array = np.array(offsets)

            # set labels whose first offset position is 0 and the second is not 0
            nr_matches: int = len(
                [elem for elem in arr_offsets if elem[0] == 0 and elem[1] != 0]
            )
            arr_tag_ids[
                (arr_offsets[:, 0] == 0) & (arr_offsets[:, 1] != 0)
            ] = tag_ids_split_into_words[index : index + nr_matches]
            index += nr_matches

            # convert
            tag_ids: List[int] = arr_tag_ids.tolist()
            all_tag_ids.append(tag_ids)

        return all_tag_ids



import os
import requests
from os.path import join, isfile

from nerblackbox.modules.datasets.formatter.base_formatter import BaseFormatter


class CoNLL2003Formatter(BaseFormatter):
    def __init__(self):
        ner_dataset = "conll2003"
        ner_tag_list = ["PER", "ORG", "LOC", "MISC"]
        super().__init__(ner_dataset, ner_tag_list)

    ####################################################################################################################
    # ABSTRACT BASE METHODS
    ####################################################################################################################
    def get_data(self, verbose: bool):
        """
        I: get data
        -----------
        :param verbose: [bool]
        :return: -
        """
        url_base = "https://raw.githubusercontent.com/patverga/torch-ner-nlp-from-scratch/master/data/conll2003/"
        targets = ["eng.train", "eng.testa", "eng.testb"]

        for target in targets:
            target_file = join(self.dataset_path, target)

            # fetch tgz from url
            if isfile(target_file):
                if verbose:
                    print(f".. file at {target_file} already exists")
            else:
                url = url_base + target
                myfile = requests.get(url, allow_redirects=True)
                open(target_file, "wb").write(myfile.content)
                if verbose:
                    print(f".. file fetched from {url} and saved at {target_file}")

    def create_ner_tag_mapping(self):
        """
        II: customize ner_training tag mapping if wanted
        -------------------------------------
        :return: ner_tag_mapping: [dict] w/ keys = tags in original data, values = tags in formatted data
        """
        return dict()

    def format_data(self):
        """
        III: format data
        ----------------
        :return: -
        """
        for phase in ["train", "val", "test"]:
            rows = self._read_original_file(phase)
            self._write_formatted_csv(phase, rows)

    def resplit_data(self, val_fraction: float):
        """
        IV: resplit data
        ----------------
        :param val_fraction: [float]
        :return: -
        """
        # train -> train
        df_train = self._read_formatted_csvs(["train"])
        self._write_final_csv("train", df_train)

        # val  -> val
        df_val = self._read_formatted_csvs(["val"])
        self._write_final_csv("val", df_val)

        # test  -> test
        df_test = self._read_formatted_csvs(["test"])
        self._write_final_csv("test", df_test)

    ####################################################################################################################
    # HELPER: READ ORIGINAL
    ####################################################################################################################
    def _read_original_file(self, phase):
        """
        III: format data
        ---------------------------------------------
        :param phase:   [str] 'train' or 'test'
        :return: _rows: [list] of [list] of [str], e.g. [[], ['Inger', 'PER'], ['säger', '0'], ..]
        """
        file_name = {
            "train": "eng.train",
            "val": "eng.testa",
            "test": "eng.testb",
        }
        file_path_original = join(self.dataset_path, file_name[phase])

        _rows = list()
        if os.path.isfile(file_path_original):
            with open(file_path_original) as f:
                for i, row in enumerate(f.readlines()):
                    _rows.append(row.strip().split())
            print(f"\n> read {file_path_original}")

        _rows = [
            [row[0], row[-1]] if (len(row) == 4 and row[0] != "-DOCSTART-") else list()
            for row in _rows
        ]

        return _rows



"""Command Line Interface of the nerblackbox package."""

import os
import subprocess
from os.path import join
import click
from typing import Dict, Any
from nerblackbox.modules.main import NerBlackBoxMain


########################################################################################################################
# CLI
########################################################################################################################
@click.group()
@click.option(
    "--data_dir", default="data", type=str, help="[str] relative path of data directory"
)
@click.option(
    "--modify/--no-modify", default=False, help="[bool] if flag=set_up_dataset"
)
@click.option(
    "--val_fraction", default=None, type=float, help="[float] if flag=set_up_dataset"
)
@click.option(
    "--verbose/--no-verbose", default=False, help="[bool] if flag=set_up_dataset"
)
@click.option("--run_name", default=None, type=str, help="[str] if flag=run_experiment")
@click.option("--device", default=None, type=str, help="[str] if flag=run_experiment")
@click.option("--fp16/--no-fp16", default=False, help="[bool] if flag=run_experiment")
@click.option("--results/--no-results", default=False, help="[bool] if flag=clear_data")
@click.pass_context
def nerbb(ctx, **kwargs_optional):
    ctx.ensure_object(dict)

    # kwargs
    kwargs = {k: v for k, v in kwargs_optional.items() if v is not None}

    # environ
    base_dir = os.getcwd()
    data_dir = kwargs.pop("data_dir")
    os.environ["BASE_DIR"] = base_dir
    os.environ["DATA_DIR"] = join(base_dir, data_dir)
    # print('BASE_DIR = ', os.environ.get('BASE_DIR'))
    # print('DATA_DIR = ', os.environ.get('DATA_DIR'))

    # context
    ctx.obj = kwargs


########################################################################################################################
# COMMANDS HELPER FUNCTION
########################################################################################################################
def _run_nerblackbox_main(_ctx_obj: Dict[str, Any], _kwargs: Dict[str, str]) -> None:
    """
    given context (_ctx_obj) and all relevant arguments (_kwargs), invoke NerBlackBoxMain
    is used by every nerbb command
    """
    kwargs = dict(**_ctx_obj, **_kwargs)

    nerblackbox_main = NerBlackBoxMain(**kwargs)
    nerblackbox_main.main()


########################################################################################################################
# COMMANDS
########################################################################################################################
@nerbb.command(name="analyze_data")
@click.pass_context
@click.argument("dataset_name")
def analyze_data(ctx, dataset_name: str):
    """analyze a dataset."""
    kwargs = {
        "flag": "analyze_data",
        "dataset_name": dataset_name,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="clear_data")
@click.pass_context
def clear_data(ctx):
    """clear data (checkpoints and optionally results)."""
    kwargs = {
        "flag": "clear_data",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="download")
@click.pass_context
def download(ctx):
    """
    download & prepare built-in datasets, prepare experiment configuration.
    needs to be called exactly once before any other CLI/API commands of the package are executed
    in case built-in datasets shall be used.
    """
    kwargs = {
        "flag": "download",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="get_experiments")
@click.pass_context
def get_experiments(ctx):
    """get overview on experiments."""
    kwargs = {
        "flag": "get_experiments",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="get_experiment_results")
@click.pass_context
@click.argument("experiment_name")
def get_experiment_results(ctx, experiment_name: str):
    """get results for a single experiment."""
    kwargs = {
        "flag": "get_experiment_results",
        "experiment_name": experiment_name,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="get_experiments_results")
@click.pass_context
def get_experiments_results(ctx):
    """get results for multiple experiments."""
    kwargs = {
        "flag": "get_experiments_results",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="init")
@click.pass_context
def init(ctx):
    """
    initialize the data_dir directory.
    needs to be called exactly once before any other CLI/API commands of the package are executed.
    """
    kwargs = {
        "flag": "init",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="mlflow")
def mlflow():
    """show detailed experiment results in mlflow (port = 5000)."""
    cd_dir = f'{join(os.environ.get("DATA_DIR"), "results")}'
    subprocess.run(f"cd {cd_dir}; mlflow ui", shell=True)


@nerbb.command(name="predict")
@click.pass_context
@click.argument("experiment_name")
@click.argument("text_input")
def predict(ctx, experiment_name: str, text_input: str):
    """predict labels for text_input using the best model of a single experiment."""
    kwargs = {
        "flag": "predict",
        "experiment_name": experiment_name,
        "text_input": text_input,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="predict_proba")
@click.pass_context
@click.argument("experiment_name")
@click.argument("text_input")
def predict_proba(ctx, experiment_name: str, text_input: str):
    """predict label probabilities for text_input using the best model of a single experiment."""
    kwargs = {
        "flag": "predict_proba",
        "experiment_name": experiment_name,
        "text_input": text_input,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="run_experiment")
@click.pass_context
@click.argument("experiment_name")
def run_experiment(ctx, experiment_name: str):
    """run a single experiment."""
    kwargs = {
        "flag": "run_experiment",
        "experiment_name": experiment_name,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="set_up_dataset")
@click.pass_context
@click.argument("dataset_name")
def set_up_dataset(ctx, dataset_name: str):
    """set up a dataset using the associated Formatter class."""
    kwargs = {
        "flag": "set_up_dataset",
        "dataset_name": dataset_name,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="show_experiment_config")
@click.pass_context
@click.argument("experiment_name")
def show_experiment_config(ctx, experiment_name: str):
    """show a single experiment configuration in detail."""
    kwargs = {
        "flag": "show_experiment_config",
        "experiment_name": experiment_name,
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="show_experiment_configs")
@click.pass_context
def show_experiment_configs(ctx):
    """show overview on all available experiment configurations."""
    kwargs = {
        "flag": "show_experiment_configs",
    }
    _run_nerblackbox_main(ctx.obj, kwargs)


@nerbb.command(name="tensorboard")
def tensorboard():
    """show detailed experiment results in tensorboard. (port = 6006)."""
    cd_dir = f'{join(os.environ.get("DATA_DIR"), "results")}'
    subprocess.run(
        f"cd {cd_dir}; tensorboard --logdir tensorboard --reload_multifile=true",
        shell=True,
    )



import os


def env_variable(key: str) -> str:
    data_dir = os.environ.get("DATA_DIR")
    assert data_dir is not None, "ERROR! DATA_DIR not found."

    env_variable_dict = {
        "DATA_DIR": data_dir,
        "DIR_DATASETS": f"{data_dir}/datasets",
        "DIR_EXPERIMENT_CONFIGS": f"{data_dir}/experiment_configs",
        "DIR_RESULTS": f"{data_dir}/results",
        "DIR_CHECKPOINTS": f"{data_dir}/results/checkpoints",
        "DIR_TENSORBOARD": f"{data_dir}/results/tensorboard",
        "DIR_MLFLOW": f"{data_dir}/results/mlruns",
        "LOG_FILE": f"{data_dir}/results/logs.log",
        "MLFLOW_FILE": f"{data_dir}/results/mlruns/mlflow_artifact.txt",
    }

    return env_variable_dict[key]



import argparse
import torch
import mlflow
import gc

import logging
import warnings

import nerblackbox.modules.ner_training.bert_ner_single as bert_ner_single
from nerblackbox.modules.utils.env_variable import env_variable
from nerblackbox.modules.experiment_config.experiment_config import ExperimentConfig

logging.basicConfig(
    level=logging.WARNING
)  # basic setting that is mainly applied to mlflow's default logging
warnings.filterwarnings("ignore")


def main(params, log_dirs):
    """
    :param params:   [argparse.Namespace] attr: experiment_name, run_name, device, fp16, experiment_run_name_nr
    :param log_dirs: [argparse.Namespace] attr: mlflow, tensorboard
    :return: -
    """
    experiment_config = ExperimentConfig(
        experiment_name=params.experiment_name,
        run_name=params.run_name,
        device=params.device,
        fp16=params.fp16,
    )
    runs_name_nr, runs_params, runs_hparams = experiment_config.parse()

    with mlflow.start_run(run_name=params.experiment_name):
        for k, v in experiment_config.get_params_and_hparams(run_name_nr=None).items():
            mlflow.log_param(k, v)

        for run_name_nr in runs_name_nr:
            # params & hparams: dict -> namespace
            params = argparse.Namespace(**runs_params[run_name_nr])
            hparams = argparse.Namespace(**runs_hparams[run_name_nr])

            # bert_ner: single run
            bert_ner_single.main(params, hparams, log_dirs, experiment=True)

            # clear gpu memory
            clear_gpu_memory(
                device=params.device, verbose=params.logging_level == "debug"
            )


def clear_gpu_memory(device, verbose: bool = False):
    """
    clear object from GPU memory
    ----------------------------
    :param device:  [torch device]
    :param verbose: [bool]
    :return: -
    """
    if device.type == "cuda":
        if verbose:
            print(torch.cuda.memory_summary(device=device))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            object_counter = 0
            for obj in gc.get_objects():
                try:
                    if obj.is_cuda:
                        object_counter += 1
                        del obj
                except:
                    pass
            gc.collect()
            torch.cuda.empty_cache()
        print(f"> cleared {object_counter} objects from GPU memory")

        if verbose:
            print(torch.cuda.memory_summary(device=device))


def _parse_args(_parser, _args):
    """
    :param _parser: [argparse ArgumentParser]
    :param _args:   [argparse arguments]
    :return _params:   [argparse.Namespace] attr: experiment_name, run_name, device, fp16
    :return _log_dirs: [argparse.Namespace] attr: mlflow, tensorboard
    """
    # parsing
    _params = None
    for group in _parser._action_groups:
        group_dict = {
            a.dest: getattr(_args, a.dest, None) for a in group._group_actions
        }
        if group.title == "args_general":
            group_dict["device"] = torch.device(
                "cuda"
                if torch.cuda.is_available() and group_dict["device"] == "gpu"
                else "cpu"
            )
            group_dict["fp16"] = (
                True
                if group_dict["fp16"] and group_dict["device"].type == "cuda"
                else False
            )
            if len(group_dict["run_name"]) == 0:
                group_dict["run_name"] = None
            _params = argparse.Namespace(**group_dict)

    # log_dirs
    _log_dirs_dict = {
        "mlflow": env_variable("DIR_MLFLOW"),
        "tensorboard": env_variable("DIR_TENSORBOARD"),
        "checkpoints": env_variable("DIR_CHECKPOINTS"),
        "log_file": env_variable("LOG_FILE"),
        "mlflow_file": env_variable("MLFLOW_FILE"),
    }
    _log_dirs = argparse.Namespace(**_log_dirs_dict)

    return _params, _log_dirs


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # params
    args_general = parser.add_argument_group("args_general")
    args_general.add_argument(
        "--experiment_name", type=str, required=True
    )  # .. logging w/ mlflow & tensorboard
    args_general.add_argument(
        "--run_name", type=str, required=True
    )  # .. logging w/ mlflow & tensorboard
    args_general.add_argument("--device", type=str, required=True)  # .. device
    args_general.add_argument("--fp16", type=int, required=True)  # .. device

    args = parser.parse_args()
    _params, _log_dirs = _parse_args(parser, args)
    main(_params, _log_dirs)



from typing import Optional, Dict
from pandas import DataFrame
from nerblackbox.modules.ner_training.ner_model_predict import NerModelPredict


class ExperimentResults:

    """class that contains results of a single experiment."""

    def __init__(
        self,
        experiment: Optional[DataFrame] = None,
        single_runs: Optional[DataFrame] = None,
        average_runs: Optional[DataFrame] = None,
        best_single_run: Optional[Dict] = None,
        best_average_run: Optional[Dict] = None,
        best_model: Optional[NerModelPredict] = None,
    ):
        """

        Args:
            experiment: overview on experiment parameters
            single_runs: overview on run parameters & single  results
            average_runs: overview on run parameters & average results
            best_single_run: overview on best run parameters & single results
            best_average_run: overview on best run parameters & average results
            best_model: best model
        """

        self.experiment = experiment
        self.single_runs = single_runs
        self.average_runs = average_runs
        self.best_single_run = best_single_run
        self.best_average_run = best_average_run
        self.best_model = best_model

    def _set_best_model(self, best_model: NerModelPredict) -> None:
        """set best model.

        Args:
            best_model: best model
        """
        self.best_model = best_model



import json


def get_ner_tag_mapping(path):
    """
    get ner_tag_mapping callable
    ----------------------------
    :param path: [str] to ner_tag_mapping json
    :return: [callable] ner_tag_mapping that applies ner_tag_mapping (loaded from json) or identical mapping
    """

    def without_bio(tag):
        """
        :param tag:                      [str], e.g. 'B-PER', 'ORG'
        :return: tag without bio prefix, [str], e.g.   'PER', 'ORG'
        """
        return tag.split("-")[-1]

    with open(path, "r") as f:
        ner_tag_mapping = json.load(f)
    return (
        lambda x: ner_tag_mapping[without_bio(x)]
        if without_bio(x) in ner_tag_mapping.keys()
        else x
    )



import os
from os.path import join
import numpy as np
import pytorch_lightning as pl
from abc import ABC, abstractmethod
import torch
from typing import List, Dict

from transformers import AdamW
from transformers import get_linear_schedule_with_warmup
from transformers import get_constant_schedule_with_warmup
from transformers import get_cosine_schedule_with_warmup
from transformers import get_cosine_with_hard_restarts_schedule_with_warmup
from transformers import AutoTokenizer

from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
    DataPreprocessor,
)
from nerblackbox.modules.utils.util_functions import split_parameters
from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation

NEWLINE_TOKENS = ["[newline]", "[NEWLINE]"]


class NerModel(pl.LightningModule, ABC):
    def __init__(self, hparams):
        """
        :param hparams: [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__()
        self.hparams = hparams

        # split up hparams
        self.params, self._hparams, self.log_dirs, self.experiment = split_parameters(
            hparams
        )

        # preparations
        self._preparations()

    ####################################################################################################################
    # Abstract Base Methods ############################################################################################
    ####################################################################################################################
    @abstractmethod
    def _preparations(self):
        """
        :created attr: default_logger    [DefaultLogger]
        :created attr: logged_metrics    [LoggedMetrics]
        :created attr: tokenizer         [transformers AutoTokenizer]
        :created attr: data_preprocessor [DataPreprocessor]
        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
        :created attr: model             [transformers AutoModelForTokenClassification]

        [only train]
        :created attr: mlflow_client          [MLflowClient]
        :created attr: epoch_metrics          [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: classification_reports [dict] w/ keys = 'val', 'test' & values = [dict]
        :created attr: pretrained_model_name  [str]
        :created attr: dataloader             [dict] w/ keys 'train', 'val', 'test' & values = [torch Dataloader]
        :created attr: optimizer              [torch optimizer]
        :created attr: scheduler              [torch LambdaLR]
        :return: -
        """
        pass

    def _preparations_data_general(self):
        """
        :created attr: pretrained_model_name  [str]
        :created attr: tokenizer              [transformers AutoTokenizer]
        :created attr: data_preprocessor      [DataPreprocessor]
        :return: -
        """
        # 1. pretrained model name
        try:
            # use transformers model
            AutoTokenizer.from_pretrained(
                self.params.pretrained_model_name, use_fast=False
            )
            self.pretrained_model_name = self.params.pretrained_model_name
        except ValueError:
            # use local model
            self.pretrained_model_name = join(
                os.environ.get("DATA_DIR"),
                "pretrained_models",
                self.params.pretrained_model_name,
            )

        # 2. tokenizer
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.pretrained_model_name,
            do_lower_case=False,
            additional_special_tokens=NEWLINE_TOKENS,
            use_fast=True,
        )  # do_lower_case needs to be False !!

        # 3. data_preprocessor
        self.data_preprocessor = DataPreprocessor(
            tokenizer=self.tokenizer,
            do_lower_case=self.params.uncased,  # can be True !!
            max_seq_length=self._hparams.max_seq_length,
            default_logger=self.default_logger,
        )

    ####################################################################################################################
    # TRAIN/VAL/TEST
    ####################################################################################################################
    ####################################################################################################################
    # FORWARD & BACKWARD PROPAGATION
    ####################################################################################################################
    def forward(self, _input_ids, _attention_mask, _segment_ids, _tag_ids):
        """
        :param _input_ids:            [torch tensor] of shape [batch_size, seq_length],
                                             e.g. 1st row = [1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]
        :param _attention_mask:       [torch tensor] of shape [batch_size, seq_length],
                                             e.g. 1st row = [1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]
        :param _segment_ids:          [torch tensor] of shape [batch_size, seq_length],
                                             e.g. 1st row = [0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]
        :param _tag_ids:              [torch tensor] of shape [batch_size, seq_length],
                                             e.g. 1st row = [1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]
        :return: _outputs: [list] of 2 elements:
                    i)  _loss: [float] cross entropy between _tag_ids & _tags_ids_predictions
                                       on non-padding tokens (i.e. where elements in _input_ids are not 0)
                    ii) _tag_ids_prediction_logits: [torch tensor] of shape [batch_size, seq_length, vocabulary_size]
        """
        return self.model(
            input_ids=_input_ids,
            attention_mask=_attention_mask,
            token_type_ids=_segment_ids,
            labels=_tag_ids,
        )

    ####################################################################################################################
    # TRAIN
    ####################################################################################################################
    def training_step(self, batch, batch_idx):
        # REQUIRED
        input_ids, attention_mask, segment_ids, tag_ids = batch
        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
        batch_train_loss = outputs[0]

        # logging
        self._write_metrics_for_tensorboard("train", {"all_loss": batch_train_loss})

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "train", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
            )

        return {"loss": batch_train_loss}

    ####################################################################################################################
    # OPTIMIZER
    ####################################################################################################################
    def configure_optimizers(self):
        # REQUIRED
        # can return multiple optimizers and learning_rate schedulers
        return self.optimizer

    def optimizer_step(
        self, current_epoch, batch_nb, optimizer, optimizer_i, second_order_closure=None
    ):
        # update params
        optimizer.step()
        optimizer.zero_grad()

        # update learning rate
        self.scheduler.step()

    ####################################################################################################################
    # VALID
    ####################################################################################################################
    def validation_step(self, batch, batch_idx):
        # OPTIONAL
        input_ids, attention_mask, segment_ids, tag_ids = batch
        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
        batch_loss, logits = outputs[:2]

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "val", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
            )

        return batch_loss, tag_ids, logits

    def validation_epoch_end(self, outputs):
        """
        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_tag_ids, batch_logits] for each batch
        :return:        [dict] w/ key 'val_loss' & value = mean batch loss of val dataset [float]
        """
        # OPTIONAL
        return self._validate_on_epoch("val", outputs=outputs)

    ####################################################################################################################
    # TEST
    ####################################################################################################################
    def test_step(self, batch, batch_idx):
        # OPTIONAL
        input_ids, attention_mask, segment_ids, tag_ids = batch
        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
        batch_loss, logits = outputs[:2]

        # debug
        if batch_idx == 0:
            self._debug_step_check(
                "test", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
            )

        return batch_loss, tag_ids, logits

    def test_epoch_end(self, outputs):
        """
        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_tag_ids, batch_logits] for each batch
        :return:        [dict] w/ key 'test_loss' & value = mean batch loss of test dataset [float]
        """
        # OPTIONAL
        return self._validate_on_epoch("test", outputs=outputs)

    ####################################################################################################################
    # DATALOADER
    ####################################################################################################################
    def train_dataloader(self):
        # REQUIRED
        return self.dataloader["train"]

    def val_dataloader(self):
        # OPTIONAL
        return self.dataloader["val"]

    def test_dataloader(self):
        # OPTIONAL
        return self.dataloader["test"]

    ####################################################################################################################
    # HELPER METHODS
    # 1. PREPARATIONS
    # 2. VALIDATE / COMPUTE METRICS
    # 3. PRINT / LOG
    ####################################################################################################################

    ####################################################################################################################
    # 1. PREPARATIONS
    ####################################################################################################################
    def _create_optimizer(
        self, learning_rate, fp16=True, no_decay=("bias", "gamma", "beta")
    ):
        """
        create optimizer with basic learning rate and L2 normalization for some parameters
        ----------------------------------------------------------------------------------
        :param learning_rate: [float] basic learning rate
        :param fp16:          [bool]
        :param no_decay:      [tuple of str] parameters that contain one of those are not subject to L2 normalization
        :return: optimizer:   [torch optimizer]
        """
        # Remove unused pooler that otherwise break Apex
        param_optimizer = list(self.model.named_parameters())
        param_optimizer = [n for n in param_optimizer if "pooler" not in n[0]]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p for n, p in param_optimizer if not any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.02,
            },
            {
                "params": [
                    p for n, p in param_optimizer if any(nd in n for nd in no_decay)
                ],
                "weight_decay": 0.0,
            },
        ]
        # print('> param_optimizer')
        # print([n for n, p in param_optimizer])
        self.default_logger.log_debug(
            "> parameters w/  weight decay:",
            len(optimizer_grouped_parameters[0]["params"]),
        )
        self.default_logger.log_debug(
            "> parameters w/o weight decay:",
            len(optimizer_grouped_parameters[1]["params"]),
        )
        if fp16:
            optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate)
            # optimizer = FusedAdam(optimizer_grouped_parameters, lr=self.learning_rate, bias_correction=False)
            # optimizer = FP16_Optimizer(optimizer, dynamic_loss_scale=True)

        else:
            optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate)
            # optimizer = FusedAdam(optimizer_grouped_parameters, lr=learning_rate)

        # optimizer = BertAdam(optimizer_grouped_parameters,lr=2e-5, warmup=.1)
        return optimizer

    def _create_scheduler(self, _lr_warmup_epochs, _lr_schedule, _lr_num_cycles=None):
        """
        create scheduler with warmup
        ----------------------------
        :param _lr_warmup_epochs:   [int]
        :param _lr_schedule:        [str], 'linear', 'constant', 'cosine', 'cosine_with_hard_resets'
        :param _lr_num_cycles:      [float, optional], e.g. 0.5, 1.0, only for cosine learning rate schedules
        :return: scheduler          [torch LambdaLR] learning rate scheduler
        """
        if _lr_schedule not in [
            "constant",
            "linear",
            "cosine",
            "cosine_with_hard_restarts",
        ]:
            raise Exception(f"lr_schedule = {_lr_schedule} not implemented.")

        num_training_steps = self._get_steps(self._hparams.max_epochs)
        num_warmup_steps = self._get_steps(_lr_warmup_epochs)

        scheduler_params = {
            "num_warmup_steps": num_warmup_steps,
            "last_epoch": -1,
        }

        if _lr_schedule == "constant":
            return get_constant_schedule_with_warmup(self.optimizer, **scheduler_params)
        else:
            scheduler_params["num_training_steps"] = num_training_steps

            if _lr_schedule == "linear":
                return get_linear_schedule_with_warmup(
                    self.optimizer, **scheduler_params
                )
            else:
                if _lr_num_cycles is not None:
                    scheduler_params[
                        "num_cycles"
                    ] = _lr_num_cycles  # else: use default values

                if _lr_schedule == "cosine":
                    scheduler_params["num_training_steps"] = num_training_steps
                    return get_cosine_schedule_with_warmup(
                        self.optimizer, **scheduler_params
                    )
                elif _lr_schedule == "cosine_with_hard_restarts":
                    scheduler_params["num_training_steps"] = num_training_steps
                    return get_cosine_with_hard_restarts_schedule_with_warmup(
                        self.optimizer, **scheduler_params
                    )
                else:
                    raise Exception(
                        "create scheduler: logic is broken."
                    )  # this should never happen

    def _get_steps(self, _num_epochs):
        """
        helper method for _create_scheduler
        gets steps = num_epochs * (number of training data samples)
        -----------------------------------------------------------------
        :param _num_epochs: [int], e.g. 10
        :return: steps:     [int], e.g. 2500 (in case of 250 training data samples)
        """
        return _num_epochs * len(self.dataloader["train"])

    def _validate_on_epoch(
        self, phase: str, outputs: List[List[torch.Tensor]]
    ) -> Dict[str, float]:
        """
        Args:
            phase:   [str], 'val', 'test'
            outputs: [list] of [lists] = [batch_loss, batch_tag_ids, batch_logits] with 3 torch tensors for each batch

        Returns:
            [dict] w/ key '<phase>_loss' & value = mean batch loss [float]
        """
        ner_model_evaluation = NerModelEvaluation(
            current_epoch=self.current_epoch,
            tag_list=self.tag_list,
            dataset_tags=self.params.dataset_tags,
            default_logger=self.default_logger,
            logged_metrics=self.logged_metrics,
        )
        epoch_metrics, classification_report, epoch_loss = ner_model_evaluation.execute(
            phase, outputs
        )
        self._log_metrics_and_classification_report(
            phase, epoch_metrics, classification_report
        )
        return {f"{phase}_loss": epoch_loss}

    def _log_metrics_and_classification_report(
        self, phase: str, epoch_metrics: Dict[str, np.array], classification_report: str
    ) -> None:
        """
        Args:
            phase:                 [str] 'val', 'test'
            epoch_metrics          [dict] w/ keys 'all_acc', 'fil_f1_micro', .. & values = [np array]
            classification_report: [str]
        """
        # tracked metrics & classification reports
        self._add_epoch_metrics(
            phase, self.current_epoch, epoch_metrics
        )  # attr: epoch_metrics

        # logging: tb
        self._write_metrics_for_tensorboard(phase, epoch_metrics)

        # logging: mlflow
        if phase == "val":
            self.mlflow_client.log_metrics(self.current_epoch, epoch_metrics)
        self.mlflow_client.log_classification_report(
            classification_report,
            overwrite=(phase == "val" and self.current_epoch == 0),
        )
        self.mlflow_client.finish_artifact_mlflow()

        # print
        self._print_metrics(phase, epoch_metrics, classification_report)

        self.default_logger.log_debug(f"--> {phase}: epoch done")

    def _add_epoch_metrics(self, phase, epoch, _epoch_metrics):
        """
        add _epoch_metrics to attribute/dict epoch_<phase>_metrics
        --------------------------------------------------------------
        :param: epoch:                [int]
        :param: _epoch_metrics:       [dict] w/ keys 'loss', 'acc', 'f1' & values = [np array]
        :changed attr: epoch_metrics: [dict] w/ keys = 'val'/'test and
                                                values = dict w/ keys = epoch [int], values = _epoch_metrics [dict]
        :return: -
        """
        self.epoch_metrics[phase][epoch] = _epoch_metrics

    def _write_metrics_for_tensorboard(self, phase, metrics):
        """
        write metrics for tensorboard
        -----------------------------
        :param phase:         [str] 'train' or 'val'
        :param metrics:       [dict] w/ keys 'loss', 'acc', 'f1_macro_all', 'f1_micro_all'
        :return: -
        """
        # tb_logs: all
        tb_logs = {
            f'{phase}/{k.split("_", 1)[0].replace("+", "P")}/{k.split("_", 1)[1]}': v
            for k, v in metrics.items()
        }

        # tb_logs: learning rate
        if phase == "train":
            tb_logs[f"{phase}/learning_rate"] = self.scheduler.get_last_lr()[0]

        # tb_logs
        self.logger.log_metrics(tb_logs, self.global_step)

    ####################################################################################################################
    # 3. PRINT / LOG
    ####################################################################################################################
    def _debug_step_check(
        self,
        phase,
        _batch,
        _outputs,
        _input_ids,
        _attention_mask,
        _segment_ids,
        _tag_ids,
    ):
        self.default_logger.log_debug(f"{phase.upper()} STEP CHECK")
        self.default_logger.log_debug(f"batch on gpu:   {_batch[0].is_cuda}")
        self.default_logger.log_debug(f"outputs on gpu: {_outputs[0].is_cuda}")
        self.default_logger.log_debug(
            f"input_ids      shape|1st row: {_input_ids.shape} | {_input_ids[0]}"
        )
        self.default_logger.log_debug(
            f"attention_mask shape|1st row: {_attention_mask.shape} | {_attention_mask[0]}"
        )
        self.default_logger.log_debug(
            f"segment_ids    shape|1st row: {_segment_ids.shape} | {_segment_ids[0]}"
        )
        self.default_logger.log_debug(
            f"tag_ids        shape|1st row: {_tag_ids.shape} | {_tag_ids[0]}"
        )

    def _print_metrics(self, phase, _metrics, _classification_reports=None):
        """
        :param phase:         [str] 'train' or 'val'
        :param _metrics:
        :param _classification_reports:
        :return:
        """
        self.default_logger.log_info("")
        self.default_logger.log_info(
            f"--- Epoch #{self.current_epoch} {phase.ljust(5).upper()} ----"
        )
        self.default_logger.log_info(
            "all loss:         {:.2f}".format(_metrics["all_loss"])
        )
        self.default_logger.log_info(
            "all acc:          {:.2f}".format(_metrics["all_acc"])
        )
        self.default_logger.log_debug(
            "all f1 (micro):   {:.2f}".format(_metrics["all_f1_micro"])
        )
        self.default_logger.log_debug(
            "all f1 (macro):   {:.2f}".format(_metrics["all_f1_macro"])
        )
        self.default_logger.log_info(
            "all  f1 (micro):   {:.2f}".format(_metrics["all_f1_micro"])
        )
        self.default_logger.log_debug(
            "all  f1 (macro):   {:.2f}".format(_metrics["all_f1_macro"])
        )
        self.default_logger.log_info(
            "fil  f1 (micro):   {:.2f}".format(_metrics["fil_f1_micro"])
        )
        self.default_logger.log_debug(
            "fil  f1 (macro):   {:.2f}".format(_metrics["fil_f1_macro"])
        )
        self.default_logger.log_info(
            "chk  f1 (micro):   {:.2f}".format(_metrics["chk_f1_micro"])
        )
        self.default_logger.log_info(f"-----------------------")
        if _classification_reports is not None:
            self.default_logger.log_debug(_classification_reports)



import torch
from typing import Dict, List
from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
    InputExample,
)

Encodings = Dict[str, torch.Tensor]
EncodingsKeys = ["input_ids", "attention_mask", "token_type_ids", "tag_ids"]
InputExamples = List[InputExample]



import os
from os.path import join
from configparser import ConfigParser
from nerblackbox.modules.utils.util_functions import get_hardcoded_parameters
from nerblackbox.modules.utils.util_functions import env_variable
from nerblackbox.modules.utils.util_functions import get_run_name, get_run_name_nr
from itertools import product
from typing import Union, Tuple, Any, Dict, Optional, List


class ExperimentConfig:
    """
    class that parses <experiment_name>.ini files
    """

    _, params, hparams, _ = get_hardcoded_parameters(keys=False)

    def __init__(self, experiment_name: str, run_name: str, device, fp16: bool):
        """
        :param experiment_name: [str],         e.g. 'exp1'
        :param run_name:        [str or None], e.g. 'runA'
        :param device:          [torch device]
        :param fp16:            [bool]
        """
        self.experiment_name = experiment_name
        self.run_name = run_name
        self.device = device
        self.fp16 = fp16

        self.config_path = join(
            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
        )
        if not os.path.isfile(self.config_path):
            raise Exception(f"config file at {self.config_path} does not exist")

        self.config_path_default = join(
            env_variable("DIR_EXPERIMENT_CONFIGS"), "default.ini"
        )
        if not os.path.isfile(self.config_path_default):
            raise Exception(
                f"default config file at {self.config_path_default} does not exist"
            )

    def get_params_and_hparams(
        self, run_name_nr: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        get dictionary of all parameters & their values that belong to
        either generally    to experiment (run_name == None)
        or     specifically to run        (run_name != None)
        --------------------------------------------------------------
        :param run_name_nr:          [str]  e.g. 'runA-1'
        :return: params_and_hparams: [dict] e.g. {'patience': 2, 'mode': 'min', ..}
        """
        _, config_dict_default = self._get_config(default=True)
        config, config_dict = self._get_config(default=False)

        if run_name_nr is None:
            params_and_hparams = config_dict_default["params"]
            params_and_hparams.update(config_dict_default["hparams"])
            params_and_hparams.update(config_dict["params"])
            params_and_hparams.update(config_dict["hparams"])
        else:
            run_name = get_run_name(run_name_nr)
            params_and_hparams = config_dict[run_name]

        return params_and_hparams

    def parse(self) -> Tuple[List[str], Dict[str, Dict], Dict[str, Dict]]:
        """
        parse <experiment_name>.ini files
        if self.run_name is specified, parse only that run. else parse all runs.
        ------------------------------------------------------------------------
        :return: _runs_name_nr [list] of [str], e.g. ['runA-1', 'runA-2', 'runB-1', 'runB-2']
                 _runs_params  [dict] w/ keys = run [str], values = params [dict],
                                      e.g. {'runA-1': {'patience': 2, 'mode': 'min', ..}, ..}
                 _runs_hparams [dict] w/ keys = run [str], values = hparams [dict],
                                      e.g. {'runA-1': {'lr_max': 2e-5, 'max_epochs': 20, ..}, ..}
        """
        _, config_dict_default = self._get_config(default=True)
        config, config_dict = self._get_config(default=False)
        if (
            "params" in config_dict.keys()
            and "multiple_runs" in config_dict["params"].keys()
        ):
            multiple_runs = config_dict["params"]["multiple_runs"]
        elif (
            "params" in config_dict_default.keys()
            and "multiple_runs" in config_dict_default["params"].keys()
        ):
            multiple_runs = config_dict_default["params"]["multiple_runs"]
        else:
            raise Exception(
                f"multiple runs is neither specified in the experiment config nor in the default config."
            )

        # _params_config & _hparams_config
        _runs_params = dict()
        _runs_hparams = dict()

        if self.run_name is None:  # multiple runs
            run_names = [
                run_name for run_name in config.sections() if run_name.startswith("run")
            ]
        else:
            run_names = [
                run_name for run_name in config.sections() if run_name == self.run_name
            ]
            assert len(run_names) == 1

        _runs_name_nr = list()
        for run_name, run_nr in product(run_names, list(range(1, multiple_runs + 1))):
            run_name_nr = get_run_name_nr(run_name, run_nr)

            # _run_params
            _run_params = {
                "experiment_name": self.experiment_name,
                "run_name": run_name,
                "run_name_nr": run_name_nr,
                "device": self.device,
                "fp16": self.fp16,
                "experiment_run_name_nr": f"{self.experiment_name}/{run_name_nr}",
            }
            _run_params.update(config_dict_default["params"])
            _run_params.update(config_dict["params"])

            # _run_hparams
            _run_hparams = dict()
            _run_hparams.update(config_dict_default["hparams"])
            _run_hparams.update(config_dict["hparams"])

            for k, v in config_dict[run_name].items():
                if k in self.params:
                    _run_params[k] = v
                elif k in self.hparams:
                    _run_hparams[k] = v
                else:
                    raise Exception(f"parameter = {k} is unknown.")

            _runs_name_nr.append(run_name_nr)
            _runs_params[run_name_nr] = _run_params
            _runs_hparams[run_name_nr] = _run_hparams

        assert set(_runs_params.keys()) == set(_runs_hparams.keys())

        return _runs_name_nr, _runs_params, _runs_hparams

    ####################################################################################################################
    # HELPER METHODS
    ####################################################################################################################
    def _get_config(self, default: bool = False) -> Tuple[ConfigParser, Dict]:
        """
        get ConfigParser instance and derive config dictionary from it
        --------------------------------------------------------------
        :param default:        [bool] if True, get default configuration instead of experiment
        :return: _config:      [ConfigParser instance]
        :return: _config_dict: [dict] w/ keys = sections [str], values = [dict] w/ params: values
        """
        _config = ConfigParser()
        _config.read(self.config_path_default if default else self.config_path)
        _config_dict: Dict[str, Dict[str, Any]] = {
            s: dict(_config.items(s)) for s in _config.sections()
        }  # {'hparams': {'monitor': 'val_loss'}}
        _config_dict = {
            s: {k: self._convert(k, v) for k, v in subdict.items()}
            for s, subdict in _config_dict.items()
        }

        # combine sections 'dataset', 'model' & 'settings' to single section 'params'
        _config_dict["params"] = dict()
        for s in ["dataset", "model", "settings"]:
            if s in _config_dict.keys():
                _config_dict["params"].update(_config_dict[s])
                _config_dict.pop(s)

        # derive uncased
        if (
            "uncased" not in _config_dict["params"].keys()
            and "pretrained_model_name" in _config_dict["params"]
        ):
            if "uncased" in _config_dict["params"]["pretrained_model_name"]:
                _config_dict["params"]["uncased"] = True
            elif "cased" in _config_dict["params"]["pretrained_model_name"]:
                _config_dict["params"]["uncased"] = False
            else:
                raise Exception(
                    "cannot derive uncased = True/False from pretrained_model_name."
                )

        return _config, _config_dict

    def _convert(
        self, _input_key: str, _input_value: str
    ) -> Union[str, int, float, bool]:
        """
        convert _input string to str/int/float/bool
        -------------------------------------------
        :param _input_key:        [str],                e.g. 'lr_schedule' or 'prune_ratio_train' or 'checkpoints'
        :param _input_value:      [str],                e.g. 'constant'    or '0.01'              or 'False'
        :return: converted_input: [str/int/float/bool], e.g. 'constant'    or  0.01               or  False
        """
        if _input_key in self.params.keys():
            convert_to = self.params[_input_key]
        elif _input_key in self.hparams.keys():
            convert_to = self.hparams[_input_key]
        else:
            raise Exception(f"_input_key = {_input_key} unknown.")

        if convert_to == "str":
            return _input_value
        elif convert_to == "int":
            return int(_input_value)
        elif convert_to == "float":
            return float(_input_value)
        elif convert_to == "bool":
            return _input_value not in ["False", "false"]
        else:
            raise Exception(f"convert_to = {convert_to} not known.")



import mlflow
from nerblackbox.modules.experiment_config.experiment_config import ExperimentConfig


class MLflowClient:
    def __init__(
        self, experiment_name, run_name, log_dirs, logged_metrics, default_logger
    ):
        """
        :param experiment_name: [str], e.g. 'Default'
        :param run_name:        [str], e.g. 'Default'
        :param log_dirs:        [Namespace], including 'mlflow_artifact' & 'default_logger_artifact'
        :param logged_metrics:  [list] of [str], e.g. ['all_precision_micro', 'all_precision_macro', ..]
        """
        self.experiment_name = experiment_name
        self.run_name = run_name
        self.log_dirs = log_dirs
        self.logged_metrics = logged_metrics  # TODO: not used !!
        self.default_logger = default_logger

    @staticmethod
    def log_params(params, hparams, experiment=False):
        """
        mlflow hyperparameter logging
        -----------------------------
        :param params:     [argparse.Namespace] attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
        :param hparams:    [argparse.Namespace] attr: batch_size, max_seq_length, max_epochs, prune_ratio_*, lr_*
        :param experiment: [bool] whether run is part of an experiment w/ multiple runs
        :return:
        """
        if experiment:
            # log only run (hyper)parameters
            experiment_config = ExperimentConfig(
                experiment_name=params.experiment_name,
                run_name=params.run_name,
                device=params.device,
                fp16=params.fp16,
            )
            for k, v in experiment_config.get_params_and_hparams(
                run_name_nr=params.run_name_nr
            ).items():
                mlflow.log_param(k, v)
        else:
            # log hardcoded set of (hyper)parameters
            if params is not None:
                # all parameters
                mlflow.log_param("parameters", vars(params))

            if hparams is not None:
                # all hyperparameters
                mlflow.log_param("hyperparameters", vars(hparams))

                # most important hyperparameters
                most_important_hyperparameters = [
                    "prune_ratio_train",
                    "prune_ratio_val",
                    "prune_ratio_test",
                    "max_epochs",
                    "lr_max",
                    "lr_schedule",
                ]
                for hyperparameter in most_important_hyperparameters:
                    mlflow.log_param(hyperparameter, vars(hparams)[hyperparameter])

    def log_metric(self, _metric, _stopped_epoch):
        mlflow.log_metric(_metric, _stopped_epoch)

    def log_metrics(self, _epoch, _epoch_val_metrics):
        """
        mlflow metrics logging
        -----------------------------
        :param: _epoch:             [int]
        :param: _epoch_val_metrics  [dict] w/ keys 'loss', 'acc', 'f1' & values = [np array]
        :return: -
        """
        mlflow.log_metric("epoch", _epoch)
        for metric in _epoch_val_metrics.keys():
            _metric = metric.replace("[", "_").replace("]", "_").replace("+", "P")
            mlflow.log_metric(_metric, _epoch_val_metrics[metric])

    def log_classification_report(self, _classification_report, overwrite=False):
        """
        log classification report
        ------------------------------------------------------------------------------------
        :param: _classification_report: [str]
        :param: overwrite: [bool] if True, overwrite existing artifact, else append
        :return: -
        """
        if overwrite:
            self._clear_artifact()
        self._log_artifact(_classification_report)

    @staticmethod
    def log_time(_time):
        mlflow.log_metric("time", _time)

    def _clear_artifact(self):
        """
        mlflow artifact logging
        -----------------------
        :return: -
        """
        with open(self.log_dirs.mlflow_file, "w") as f:
            f.write(" ")

    def _log_artifact(self, content):
        """
        mlflow artifact logging
        -----------------------
        :param content: [str]
        :return: -
        """
        with open(self.log_dirs.mlflow_file, "a") as f:
            f.write(content + "\n")

    def finish_artifact_mlflow(self):
        # mlflow
        mlflow.log_artifact(self.log_dirs.mlflow_file)
        self.default_logger.log_debug(f"mlflow file at {self.log_dirs.mlflow_file}")

    def finish_artifact_logger(self):
        # default logger
        mlflow.log_artifact(self.log_dirs.log_file)
        self.default_logger.log_debug(f"log file at {self.log_dirs.log_file}")



import argparse
from nerblackbox.modules.datasets.formatter.auto_formatter import AutoFormatter
import mlflow


def main(args):
    """
    - V: analyze data
      . read reshaped text files
      . analyze data and output class distribution
    --------------------------------------------------------------------------------
    :return: -
    """
    with mlflow.start_run(run_name="Default"):
        formatter = AutoFormatter.for_dataset(args.ner_dataset)
        formatter.analyze_data()  # IV: analyze data
        formatter.plot_data()  # IV: analyze data


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ner_dataset", required=True, type=str, help="e.g. swedish_ner_corpus"
    )
    parser.add_argument("--verbose", type=bool, default=False)
    _args = parser.parse_args()

    main(_args)



import json

import numpy as np
from transformers import AutoModelForTokenClassification
from argparse import Namespace
from torch.nn.functional import softmax
from typing import List, Union, Tuple, Any, Dict

from nerblackbox.modules.ner_training.ner_model import NerModel


class NerModelPredict(NerModel):
    """
    class that predicts tags for given text
    """

    @classmethod
    def load_from_checkpoint(
        cls,
        checkpoint_path: str,
    ) -> "NerModelPredict":
        """load model in inference mode from checkpoint_path

        Args:
            checkpoint_path: path to checkpoint

        Returns:
            model loaded from checkpoint
        """
        model = super().load_from_checkpoint(
            checkpoint_path, map_location=None, tags_csv=None
        )
        model.freeze()  # for inference mode
        return model

    def __init__(self, hparams: Namespace):
        """
        Args:
            hparams: attr experiment_name, run_name, pretrained_model_name, dataset_name, ..
        """
        super().__init__(hparams)

    ####################################################################################################################
    # Abstract Base Methods ############################################################################################
    ####################################################################################################################
    def _preparations(self):
        """
        :created attr: default_logger    [DefaultLogger]
        :created attr: logged_metrics    [LoggedMetrics]
        :created attr: tokenizer         [transformers AutoTokenizer]
        :created attr: data_preprocessor [DataPreprocessor]
        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
        :created attr: model             [transformers AutoModelForTokenClassification]
        :return: -
        """
        # predict
        self._preparations_predict()  # attr: default_logger
        self._preparations_data_general()  # attr: tokenizer, data_preprocessor
        self._preparations_data_predict()  # attr: tag_list, model

    def _preparations_predict(self):
        """
        :created attr: default_logger         [None]
        :return: -
        """
        self.default_logger = None

    def _preparations_data_predict(self):
        """
        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
        :created attr: model             [transformers AutoModelForTokenClassification]
        :return: -
        """
        # tag_list
        self.tag_list = json.loads(self.hparams.tag_list)

        # model
        self.model = AutoModelForTokenClassification.from_pretrained(
            self.pretrained_model_name,
            num_labels=len(self.tag_list),
            return_dict=False,
        )
        self.model.resize_token_embeddings(
            len(self.tokenizer)
        )  # due to addtional_special_tokens = NEWLINE_TOKENS

    ####################################################################################################################
    # PREDICT
    ####################################################################################################################
    def predict(self, examples: Union[str, List[str]]) -> List[Namespace]:
        """predict tags

        Args:
            examples: e.g. ["example 1", "example 2"]

        Returns:
            predictions: with .internal [list] of (word, tag) tuples \
                         and  .external [list] of (word, tag) tuples
        """
        return self._predict(examples, proba=False)

    def predict_proba(self, examples: Union[str, List[str]]) -> List[Namespace]:
        """predict probabilities for tags

        Args:
            examples: e.g. ["example 1", "example 2"]
        Returns:
            predictions: with .internal [list] of (word, proba_dist) tuples \
                         and  .external [list] of (word, proba_dist) tuples \
                         where proba_dist = [dict] that maps self.tag_list to probabilities
        """
        return self._predict(examples, proba=True)

    def _predict(
        self, examples: Union[str, List[str]], proba: bool = False
    ) -> List[Namespace]:
        """predict tags or probabilities for tags

        Args:
            examples: e.g. ["example 1", "example 2"]
            proba: predict probabilities instead of labels

        Returns:
            predictions: with .internal [list] of (word, tag / proba_dist) tuples
                         and  .external [list] of dict w/ keys = char_start, char_end, word, tag / proba_dist
        """
        if isinstance(examples, str):
            examples = [examples]

        predictions = list()
        for example in examples:
            predict_dataloader = self._get_predict_dataloader([example])

            if self.data_preprocessor.do_lower_case:
                example = example.lower()

            example_tokens = list()
            example_token_predictions = list()
            ######################################
            # 1 example, individual chunks, tokens
            ######################################
            for sample in predict_dataloader:
                token_tensors, tokens = self._predict_on_tokens(sample)
                if proba is False:
                    token_predictions = self._turn_tensors_into_tags(token_tensors)
                else:
                    token_predictions = (
                        self._turn_tensors_into_tag_probability_distributions(
                            token_tensors
                        )
                    )
                example_tokens.extend(tokens)
                example_token_predictions.extend(token_predictions)

            ######################################
            # 1 example, merged chunks, tokens -> words
            ######################################
            example_word_predictions = self._get_tags_on_words_between_special_tokens(
                example_tokens, example_token_predictions
            )

            example_word_predictions["external"] = self._restore_unknown_tokens(
                example_word_predictions["external"], example
            )

            prediction = self._namespace_prediction(
                example_word_predictions,
            )

            predictions.append(prediction)

        ######################################
        # all examples, merged chunks, words
        ######################################
        return predictions

    ####################################################################################################################
    # PREDICT HELPER METHODS
    ####################################################################################################################
    def _get_predict_dataloader(self, examples):
        """
        :param examples:            [list] of [str]
        :return: predict_dataloader [torch dataloader]
        """
        # input_examples
        input_examples = self.data_preprocessor.get_input_examples_predict(
            examples=examples,
        )

        # dataloader
        dataloader = self.data_preprocessor.to_dataloader(
            input_examples, self.tag_list, batch_size=1
        )

        return dataloader["predict"]

    def _predict_on_tokens(self, sample):
        """
        :param sample: [list] w/ 4 tensors: input_ids, attention_mask, segment_ids, label_ids
        :return: output_token_tensors [list] of [torch tensor] of shape [1, #tags]
        :return: tokens               [list] of [str]
        """
        (
            input_ids,  # shape: [1, seq_length]
            attention_mask,  # shape: [1, seq_length]
            segment_ids,  # shape: [1, seq_length]
            label_ids,  # shape: [1, seq_length]
        ) = sample

        output = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=segment_ids,
        )  # shape: [1 (=#examples), 1 (=#batch_size), seq_length, #tags]

        output_token_tensors = [
            output[0][0][i]  # .detach().numpy()
            for i in range(self._hparams.max_seq_length)
        ]  # shape: [seq_length, #tags]

        tokens = self.tokenizer.convert_ids_to_tokens(input_ids[0])

        return output_token_tensors, tokens

    def _turn_tensors_into_tags(self, output_token_tensors):
        """
        :param output_token_tensors       [list] of [torch tensor] of shape [1, #tags]
        :return: output_token_predictions [list] of [str]
        """
        return [
            self.tag_list[np.argmax(output_token_tensors[i].detach().numpy())]
            for i in range(self._hparams.max_seq_length)
        ]

    def _turn_tensors_into_tag_probability_distributions(self, output_token_tensors):
        """
        :param output_token_tensors       [list] of [torch tensor] of shape [1, #tags]
        :return: output_token_predictions [list] of [prob dist], i.e. dict that maps tags to probabilities
        """

        probability_distributions = [
            softmax(output_token_tensors[i], dim=0)
            for i in range(self._hparams.max_seq_length)
        ]

        tag_probability_distribution = [
            {
                self.tag_list[j]: float(
                    probability_distributions[i][j].detach().numpy()
                )
                for j in range(len(self.tag_list))
            }
            for i in range(self._hparams.max_seq_length)
        ]

        return tag_probability_distribution

    @staticmethod
    def _get_tags_on_words_between_special_tokens(
        tokens: List[str], example_token_predictions: List[Any]
    ):
        """
        Args:
            tokens:                    [list] of [str]
            example_token_predictions  [list] of [str] or [prob dist]

        Returns:
            example_word_predictions   [list] of [str] or [prob dist]
        """
        # predict tags on words between [CLS] and [SEP]
        example_word_predictions: Dict[str, List[Any]] = {
            "internal": list(),
            "external": list(),
        }
        for token, example_token_prediction in zip(tokens, example_token_predictions):
            if token not in ["[CLS]", "[SEP]", "[PAD]"]:
                example_word_predictions["internal"].append(
                    [token, example_token_prediction]
                )
                if not token.startswith("##"):
                    example_word_predictions["external"].append(
                        [token, example_token_prediction]
                    )
                else:
                    example_word_predictions["external"][-1][0] += token.strip("##")

        for field in ["internal", "external"]:
            example_word_predictions[field] = [
                tuple(elem) for elem in example_word_predictions[field]
            ]

        return example_word_predictions

    @staticmethod
    def _restore_unknown_tokens(
        example_word_predictions_external: List[Tuple[str, str]],
        _example: str,
        verbose: bool = True,
    ) -> List[Dict[str, str]]:
        """
        - replace "[UNK]" tokens by the original token
        - enrich tokens with char_start & char_end

        Args:
            example_word_predictions_external: e.g. [('example', 'O'), ('of', 'O), ('author', 'PERSON'), ..]
            _example: 'example of author'

        Returns:
            example_word_predictions_external: e.g. [('example', 'O'), ('of', 'O), ('author', 'PERSON'), ..]
        """
        _predictions_external = list()

        # 1. get margins of known tokens
        token_char_margins: List[Tuple[Any, ...]] = list()
        char_start = 0
        for token, _ in example_word_predictions_external:
            if token != "[UNK]":
                try:
                    char_start = _example.index(token, char_start)
                    token_char_margins.append((char_start, char_start + len(token)))
                    char_start += len(token)
                except ValueError:
                    print(f"! token = {token} not found in example[{char_start}:]")
                    token_char_margins.append((None, None))
            else:
                token_char_margins.append((None, None))

        # 2. restore unknown tokens
        for i, (token, tag) in enumerate(example_word_predictions_external):
            if (
                token_char_margins[i][0] is not None
                and token_char_margins[i][1] is not None
            ):
                _predictions_external.append(
                    {
                        "char_start": str(token_char_margins[i][0]),
                        "char_end": str(token_char_margins[i][1]),
                        "token": token,
                        "tag": tag,
                    }
                )
            else:
                for j in range(2):
                    assert (
                        token_char_margins[i][j] is None
                    ), f"ERROR! token_char_margin[{i}][{j}] is not None for token = {token}"

                char_start_margin, char_end_margin = None, None
                for k in range(1, 10):
                    if i - k < 0:
                        char_start_margin = 0
                        break
                    elif token_char_margins[i - k][1] is not None:
                        char_start_margin = token_char_margins[i - k][1]
                        break
                for k in range(1, 10):
                    if i + k >= len(token_char_margins):
                        char_end_margin = len(_example)
                        break
                    elif token_char_margins[i + k][0] is not None:
                        char_end_margin = token_char_margins[i + k][0]
                        break
                assert (
                    char_start_margin is not None
                ), f"ERROR! could not find char_start_margin"
                assert (
                    char_end_margin is not None
                ), f"ERROR! could not find char_end_margin"

                new_token = _example[char_start_margin:char_end_margin].strip()

                if len(new_token):
                    char_start = _example.index(new_token, char_start_margin)
                    char_end = char_start + len(new_token)
                    if verbose:
                        print(
                            f"! restored unknown token = {new_token} between "
                            f"char_start = {char_start_margin}, "
                            f"char_end = {char_end_margin}"
                        )
                    _predictions_external.append(
                        {
                            "char_start": str(char_start),
                            "char_end": str(char_end),
                            "token": new_token,
                            "tag": tag,
                        }
                    )
                else:
                    print(
                        f"! dropped unknown empty token between "
                        f"char_start = {char_start_margin}, "
                        f"char_end = {char_end_margin}"
                    )

        return _predictions_external

    @staticmethod
    def _namespace_prediction(example_word_predictions):
        """
        :param example_word_predictions   [dict] w/ keys = "internal", "external"
                                                   & values = [list] of [str] or [prob dist]
        :return: prediction              [Namespace] w/ attributes = 'internal' | 'external'
                                                     & values = [list] of [tuples] (word, tag / tag prob dist)
        """
        prediction = Namespace(**example_word_predictions)
        return prediction



import pytest
import numpy as np
import torch
from typing import Dict, List, Union
from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation
from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics


class TestNerModelEvaluation:

    ner_model_evaluation = NerModelEvaluation(
        current_epoch=2,
        tag_list=["O", "PER", "ORG"],
        dataset_tags=["O", "PER", "ORG", "MISC"],
        default_logger=None,
        logged_metrics=LoggedMetrics(),
    )

    ####################################################################################################################
    # TEST #############################################################################################################
    ####################################################################################################################
    @pytest.mark.parametrize(
        "outputs, " "np_batch, " "np_epoch",
        [
            (
                [
                    [
                        torch.tensor([1.0, 2.0]),
                        torch.tensor([3, 4]),
                        torch.tensor([[9.0, 1.2, 3.3], [0.7, 0.6, 5.6]]),
                    ],
                    [
                        torch.tensor([2.0, 3.0]),
                        torch.tensor([5, 6]),
                        torch.tensor([[0.9, 2.1, 3.3], [7.0, 6.0, 6.5]]),
                    ],
                ],
                {
                    "loss": [
                        np.array([1.0, 2.0], dtype=np.float32),
                        np.array([2.0, 3.0], dtype=np.float32),
                    ],
                    "tag_ids": [
                        np.array([3, 4], dtype=np.float32),
                        np.array([5, 6], dtype=np.float32),
                    ],
                    "logits": [
                        np.array([[9.0, 1.2, 3.3], [0.7, 0.6, 5.6]], dtype=np.float32),
                        np.array([[0.9, 2.1, 3.3], [7.0, 6.0, 6.5]], dtype=np.float32),
                    ],
                },
                {
                    "loss": 2.0,
                    "tag_ids": np.array([3, 4, 5, 6], dtype=np.float32),
                    "logits": np.array(
                        [
                            [9.0, 1.2, 3.3],
                            [0.7, 0.6, 5.6],
                            [0.9, 2.1, 3.3],
                            [7.0, 6.0, 6.5],
                        ],
                        dtype=np.float32,
                    ),
                },
            ),
        ],
    )
    def test_convert_and_combine(
        self,
        outputs: List[List[torch.Tensor]],
        np_batch: Dict[str, List[np.array]],
        np_epoch: Dict[str, Union[np.number, np.array]],
    ) -> None:
        # 1. output -> np_batch
        test_np_batch = self.ner_model_evaluation._convert_output_to_np_batch(outputs)
        for key in np_batch.keys():
            for _test, _true in zip(test_np_batch[key], np_batch[key]):
                assert np.array_equal(
                    _test, _true
                ), f"test_np_batch[{key}] = {_test} != {_true}"

        # 2. np_batch -> np_epoch
        test_np_epoch = self.ner_model_evaluation._combine_np_batch_to_np_epoch(
            np_batch
        )
        for key in np_epoch.keys():
            if key == "loss":
                assert (
                    test_np_epoch[key] == np_epoch[key]
                ), f"test_np_epoch[{key}] = {test_np_epoch[key]} != {np_epoch[key]}"
            else:
                for _test, _true in zip(test_np_epoch[key], np_epoch[key]):
                    assert np.array_equal(
                        _test, _true
                    ), f"test_np_epoch[{key}] = {_test} != {_true}"

    @pytest.mark.parametrize(
        "_np_tag_ids, " "_np_logits, " "true_flat, " "pred_flat",
        [
            (
                np.array([[0, 1], [1, 0]]),
                np.array([[[0.8, 0.2], [0.8, 0.2]], [[0.3, 0.7], [0.3, 0.7]]]),
                np.array([0, 1, 1, 0]),
                np.array([0, 0, 1, 1]),
            ),
        ],
    )
    def test_reduce_and_flatten(
        self,
        _np_tag_ids: np.array,
        _np_logits: np.array,
        true_flat: np.array,
        pred_flat: np.array,
    ) -> None:
        test_true_flat, test_pred_flat = self.ner_model_evaluation._reduce_and_flatten(
            _np_tag_ids, _np_logits
        )
        assert list(test_true_flat) == list(
            true_flat
        ), f"true_flat: test = {test_true_flat} != {true_flat} = true"
        assert list(test_pred_flat) == list(
            pred_flat
        ), f"pred_flat: test = {test_true_flat} != {pred_flat} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_np_tag_ids, " "tags",
        [
            (
                np.array([0, 1, 2, 1]),
                np.array(["O", "PER", "ORG", "PER"]),
            ),
            (
                np.array([-100, 1, 2, -100]),
                np.array(["[S]", "PER", "ORG", "[S]"]),
            ),
        ],
    )
    def test_convert_tag_ids_to_tags(
        self,
        _np_tag_ids: np.array,
        tags: np.array,
    ) -> None:
        test_tags = self.ner_model_evaluation._convert_tag_ids_to_tags(_np_tag_ids)
        assert list(test_tags) == list(tags), f"test = {test_tags} != {tags} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_tags, " "tags_new",
        [
            (
                {
                    "true": np.array(["[S]", "PER", "ORG", "[S]"]),
                    "pred": np.array(["PER", "ORG", "O", "O"]),
                },
                {
                    "true": np.array(["PER", "ORG"]),
                    "pred": np.array(["ORG", "O"]),
                },
            ),
        ],
    )
    def test_get_rid_of_special_tag_occurrences(
        self,
        _tags: Dict[str, np.array],
        tags_new: Dict[str, np.array],
    ) -> None:
        test_tags_new = self.ner_model_evaluation._get_rid_of_special_tag_occurrences(
            _tags
        )
        for k in tags_new.keys():
            assert list(test_tags_new[k]) == list(
                tags_new[k]
            ), f"key = {k}: test = {test_tags_new[k]} != {tags_new[k]} = true"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_tag_subset, " "_filtered_tags",
        [
            ("all", ["O", "PER", "ORG"]),
            ("fil", ["PER", "ORG"]),
            ("chk", ["PER", "ORG"]),
            ("O", ["O"]),
            ("PER", ["PER"]),
            ("ORG", ["ORG"]),
        ],
    )
    def test_get_filtered_tags(
        self,
        _tag_subset: str,
        _filtered_tags: List[str],
    ) -> None:
        test_filtered_tags = self.ner_model_evaluation._get_filtered_tags(_tag_subset)
        assert (
            test_filtered_tags == _filtered_tags
        ), f"test_filtered_tags = {test_filtered_tags} != {_filtered_tags}"

    ####################################################################################################################
    @pytest.mark.parametrize(
        "_tags, " "_phase, " "tag_subset, " "metrics",
        [
            (
                {
                    "true": np.array(["O", "PER", "ORG", "PER"]),
                    "pred": np.array(["O", "PER", "ORG", "PER"]),
                },
                "val",
                "fil",
                {
                    "fil_recall_micro": 1.0,
                    "fil_recall_macro": 1.0,
                    "fil_precision_micro": 1.0,
                    "fil_precision_macro": 1.0,
                    "fil_f1_micro": 1.0,
                    "fil_f1_macro": 1.0,
                },
            ),
        ],
    )
    def test_compute_metrics_for_tags_subset(
        self,
        _tags: Dict[str, np.array],
        _phase: str,
        tag_subset: str,
        metrics: Dict[str, float],
    ) -> None:
        test_metrics = self.ner_model_evaluation._compute_metrics_for_tags_subset(
            _tags, _phase, tag_subset
        )
        for k, v in test_metrics.items():
            assert (
                test_metrics[k] == metrics[k]
            ), f"test_metrics[{k}] = {test_metrics[k]} != {metrics[k]}"


if __name__ == "__main__":
    test = TestNerModelEvaluation()
    test.test_convert_and_combine()
    test.test_reduce_and_flatten()
    test.test_convert_tag_ids_to_tags()
    test.test_get_rid_of_special_tag_occurrences()
    test.test_get_filtered_tags()
    test.test_compute_metrics_for_tags_subset()



from typing import Optional
from pandas import DataFrame


class ExperimentsResults:

    """class that contains results of multiple experiments."""

    def __init__(
        self,
        best_single_runs: Optional[DataFrame] = None,
        best_average_runs: Optional[DataFrame] = None,
    ):
        """
        Args:
            best_single_runs: overview on best single runs
            best_average_runs: overview on best average runs
        """

        self.best_single_runs = best_single_runs
        self.best_average_runs = best_average_runs



import logging

import os
from os.path import join

BASE_DIR = os.environ.get("BASE_DIR")


class DefaultLogger:

    custom_datefmt = "%Y/%m/%d %H:%M:%S"
    custom_format = "%(asctime)s - %(name)s - %(levelname)s -  %(message)s"
    assert BASE_DIR is not None, "ERROR! BASE_DIR not found."
    default_log_file = join(BASE_DIR, "logs.log")

    def __init__(self, path, log_file=None, level=None, mode="a"):
        """
        :param path:     [str] of python module that uses DefaultLogger, e.g. [..]/module.py
        :param log_file: [str] of log file, e.g. [..]/logs.log
        :param level:    [str] 'debug', 'info' or 'warning (minimum logging level)
        :param mode:     [str] 'w' (write) or 'a' (append)
        :created attr: logger [python logging logger]
        """

        name = path.split("/")[-1]
        self.logger = logging.getLogger(name)
        self.logger.propagate = (
            level == "debug"
        )  # only shows console output if level is 'debug'

        # log file
        self.filename = log_file if log_file else self.default_log_file

        # level
        self.level = level.upper() if level else None
        if self.level:
            self.logger.setLevel(self.level)

        # add file handler
        if (
            len(self.logger.handlers) == 0
        ):  # avoid adding same file handler multiple times
            f_handler = logging.FileHandler(filename=self.filename, mode=mode)
            f_handler.setFormatter(
                logging.Formatter(self.custom_format, datefmt=self.custom_datefmt)
            )
            if self.level:
                f_handler.setLevel(self.level)
            self.logger.addHandler(f_handler)

        if mode == "w":
            self.log_debug(f"ACTIVATE FILE LOGGER {name} w/ PATH = {self.filename}")
        elif mode == "a":
            self.log_debug(f"APPEND TO FILE LOGGER {name} w/ PATH = {self.filename}")

    def clear(self):
        """
        clear file at log_file
        """
        with open(self.filename, "w") as f:
            f.write("")

    def log_debug(self, *args):
        """
        debug: log & print
        """
        msg = " ".join([str(elem) for elem in args])
        self.logger.debug(msg)  # log
        if self.level in ["DEBUG"]:
            print("DEBUG -----", msg)  # print

    def log_info(self, *args):
        """
        info: log & print
        """
        msg = " ".join([str(elem) for elem in args])
        self.logger.info(msg)  # log
        if self.level in ["DEBUG", "INFO"]:
            print("INFO ------", msg)  # print

    def log_warning(self, *args):
        """
        warning: log & print
        """
        msg = " ".join([str(elem) for elem in args])
        self.logger.warning(msg)  # log
        if self.level in ["DEBUG", "INFO", "WARNING"]:
            print("WARNING ---", msg)  # print
